"""Python Enumerations"""

import sys as _sys

__all__ = ['Enum', 'IntEnum', 'unique']

pyver = float('%s.%s' % _sys.version_info[:2])

try:
    any
except NameError:
    def any(iterable):
        for element in iterable:
            if element:
                return True
        return False


class _RouteClassAttributeToGetattr(object):
    """Route attribute access on a class to __getattr__.

    This is a descriptor, used to define attributes that act differently when
    accessed through an instance and through a class.  Instance access remains
    normal, but access to an attribute through a class will be routed to the
    class's __getattr__ method; this is done by raising AttributeError.

    """
    def __init__(self, fget=None):
        self.fget = fget

    def __get__(self, instance, ownerclass=None):
        if instance is None:
            raise AttributeError()
        return self.fget(instance)

    def __set__(self, instance, value):
        raise AttributeError("can't set attribute")

    def __delete__(self, instance):
        raise AttributeError("can't delete attribute")


def _is_descriptor(obj):
    """Returns True if obj is a descriptor, False otherwise."""
    return (
            hasattr(obj, '__get__') or
            hasattr(obj, '__set__') or
            hasattr(obj, '__delete__'))


def _is_dunder(name):
    """Returns True if a __dunder__ name, False otherwise."""
    return (name[:2] == name[-2:] == '__' and
            name[2:3] != '_' and
            name[-3:-2] != '_' and
            len(name) > 4)


def _is_sunder(name):
    """Returns True if a _sunder_ name, False otherwise."""
    return (name[0] == name[-1] == '_' and 
            name[1:2] != '_' and
            name[-2:-1] != '_' and
            len(name) > 2)


def _make_class_unpicklable(cls):
    """Make the given class un-picklable."""
    def _break_on_call_reduce(self):
        raise TypeError('%r cannot be pickled' % self)
    cls.__reduce__ = _break_on_call_reduce
    cls.__module__ = '<unknown>'


class _EnumDict(dict):
    """Track enum member order and ensure member names are not reused.

    EnumMeta will use the names found in self._member_names as the
    enumeration member names.

    """
    def __init__(self):
        super(_EnumDict, self).__init__()
        self._member_names = []

    def __setitem__(self, key, value):
        """Changes anything not dundered or not a descriptor.

        If a descriptor is added with the same name as an enum member, the name
        is removed from _member_names (this may leave a hole in the numerical
        sequence of values).

        If an enum member name is used twice, an error is raised; duplicate
        values are not checked for.

        Single underscore (sunder) names are reserved.

        Note:   in 3.x __order__ is simply discarded as a not necessary piece
                leftover from 2.x

        """
        if pyver >= 3.0 and key == '__order__':
                return
        if _is_sunder(key):
            raise ValueError('_names_ are reserved for future Enum use')
        elif _is_dunder(key):
            pass
        elif key in self._member_names:
            # descriptor overwriting an enum?
            raise TypeError('Attempted to reuse key: %r' % key)
        elif not _is_descriptor(value):
            if key in self:
                # enum overwriting a descriptor?
                raise TypeError('Key already defined as: %r' % self[key])
            self._member_names.append(key)
        super(_EnumDict, self).__setitem__(key, value)


# Dummy value for Enum as EnumMeta explicitly checks for it, but of course until
# EnumMeta finishes running the first time the Enum class doesn't exist.  This
# is also why there are checks in EnumMeta like `if Enum is not None`
Enum = None


class EnumMeta(type):
    """Metaclass for Enum"""
    @classmethod
    def __prepare__(metacls, cls, bases):
        return _EnumDict()

    def __new__(metacls, cls, bases, classdict):
        # an Enum class is final once enumeration items have been defined; it
        # cannot be mixed with other types (int, float, etc.) if it has an
        # inherited __new__ unless a new __new__ is defined (or the resulting
        # class will fail).
        if type(classdict) is dict:
            original_dict = classdict
            classdict = _EnumDict()
            for k, v in original_dict.items():
                classdict[k] = v

        member_type, first_enum = metacls._get_mixins_(bases)
        #if member_type is object:
        #    use_args = False
        #else:
        #    use_args = True
        __new__, save_new, use_args = metacls._find_new_(classdict, member_type,
                                                        first_enum)
        # save enum items into separate mapping so they don't get baked into
        # the new class
        members = dict((k, classdict[k]) for k in classdict._member_names)
        for name in classdict._member_names:
            del classdict[name]

        # py2 support for definition order
        __order__ = classdict.get('__order__')
        if __order__ is None:
            __order__ = classdict._member_names
            if pyver < 3.0:
                order_specified = False
            else:
                order_specified = True
        else:
            del classdict['__order__']
            order_specified = True
            if pyver < 3.0:
                __order__ = __order__.replace(',', ' ').split()
                aliases = [name for name in members if name not in __order__]
                __order__ += aliases

        # check for illegal enum names (any others?)
        invalid_names = set(members) & set(['mro'])
        if invalid_names:
            raise ValueError('Invalid enum member name(s): %s' % (
                ', '.join(invalid_names), ))

        # create our new Enum type
        enum_class = super(EnumMeta, metacls).__new__(metacls, cls, bases, classdict)
        enum_class._member_names_ = []               # names in random order
        enum_class._member_map_ = {}                 # name->value map
        enum_class._member_type_ = member_type

        # Reverse value->name map for hashable values.
        enum_class._value2member_map_ = {}

        # check for a __getnewargs__, and if not present sabotage
        # pickling, since it won't work anyway
        if (member_type is not object and
            member_type.__dict__.get('__getnewargs__') is None
            ):
            _make_class_unpicklable(enum_class)

        # instantiate them, checking for duplicates as we go
        # we instantiate first instead of checking for duplicates first in case
        # a custom __new__ is doing something funky with the values -- such as
        # auto-numbering ;)
        if __new__ is None:
            __new__ = enum_class.__new__
        for member_name in __order__:
            value = members[member_name]
            if not isinstance(value, tuple):
                args = (value, )
            else:
                args = value
            if member_type is tuple:   # special case for tuple enums
                args = (args, )     # wrap it one more time
            if not use_args or not args:
                enum_member = __new__(enum_class)
                if not hasattr(enum_member, '_value_'):
                    enum_member._value_ = value
            else:
                enum_member = __new__(enum_class, *args)
                if not hasattr(enum_member, '_value_'):
                    enum_member._value_ = member_type(*args)
            value = enum_member._value_
            enum_member._name_ = member_name
            enum_member.__objclass__ = enum_class
            enum_member.__init__(*args)
            # If another member with the same value was already defined, the
            # new member becomes an alias to the existing one.
            for name, canonical_member in enum_class._member_map_.items():
                if canonical_member.value == enum_member._value_:
                    enum_member = canonical_member
                    break
            else:
                # Aliases don't appear in member names (only in __members__).
                enum_class._member_names_.append(member_name)
            enum_class._member_map_[member_name] = enum_member
            try:
                # This may fail if value is not hashable. We can't add the value
                # to the map, and by-value lookups for this value will be
                # linear.
                enum_class._value2member_map_[value] = enum_member
            except TypeError:
                pass

        # in Python2.x we cannot know definition order, so go with value order
        # unless __order__ was specified in the class definition
        if not order_specified:
            enum_class._member_names_ = [
                e[0] for e in sorted(
                [(name, enum_class._member_map_[name]) for name in enum_class._member_names_],
                 key=lambda t: t[1]._value_
                        )]

        # double check that repr and friends are not the mixin's or various
        # things break (such as pickle)
        if Enum is not None:
            setattr(enum_class, '__getnewargs__', Enum.__getnewargs__)
        for name in ('__repr__', '__str__', '__format__'):
            class_method = getattr(enum_class, name)
            obj_method = getattr(member_type, name, None)
            enum_method = getattr(first_enum, name, None)
            if obj_method is not None and obj_method is class_method:
                setattr(enum_class, name, enum_method)

        # method resolution and int's are not playing nice
        # Python's less than 2.6 use __cmp__

        if pyver < 2.6:

            if issubclass(enum_class, int):
                setattr(enum_class, '__cmp__', getattr(int, '__cmp__'))

        elif pyver < 3.0:

            if issubclass(enum_class, int):
                for method in (
                        '__le__',
                        '__lt__',
                        '__gt__',
                        '__ge__',
                        '__eq__',
                        '__ne__',
                        '__hash__',
                        ):
                    setattr(enum_class, method, getattr(int, method))

        # replace any other __new__ with our own (as long as Enum is not None,
        # anyway) -- again, this is to support pickle
        if Enum is not None:
            # if the user defined their own __new__, save it before it gets
            # clobbered in case they subclass later
            if save_new:
                setattr(enum_class, '__member_new__', enum_class.__dict__['__new__'])
            setattr(enum_class, '__new__', Enum.__dict__['__new__'])
        return enum_class

    def __call__(cls, value, names=None, module=None, type=None):
        """Either returns an existing member, or creates a new enum class.

        This method is used both when an enum class is given a value to match
        to an enumeration member (i.e. Color(3)) and for the functional API
        (i.e. Color = Enum('Color', names='red green blue')).

        When used for the functional API: `module`, if set, will be stored in
        the new class' __module__ attribute; `type`, if set, will be mixed in
        as the first base class.

        Note: if `module` is not set this routine will attempt to discover the
        calling module by walking the frame stack; if this is unsuccessful
        the resulting class will not be pickleable.

        """
        if names is None:  # simple value lookup
            return cls.__new__(cls, value)
        # otherwise, functional API: we're creating a new Enum type
        return cls._create_(value, names, module=module, type=type)

    def __contains__(cls, member):
        return isinstance(member, cls) and member.name in cls._member_map_

    def __delattr__(cls, attr):
        # nicer error message when someone tries to delete an attribute
        # (see issue19025).
        if attr in cls._member_map_:
            raise AttributeError(
                    "%s: cannot delete Enum member." % cls.__name__)
        super(EnumMeta, cls).__delattr__(attr)

    def __dir__(self):
        return (['__class__', '__doc__', '__members__', '__module__'] +
                self._member_names_)

    @property
    def __members__(cls):
        """Returns a mapping of member name->value.

        This mapping lists all enum members, including aliases. Note that this
        is a copy of the internal mapping.

        """
        return cls._member_map_.copy()

    def __getattr__(cls, name):
        """Return the enum member matching `name`

        We use __getattr__ instead of descriptors or inserting into the enum
        class' __dict__ in order to support `name` and `value` being both
        properties for enum members (which live in the class' __dict__) and
        enum members themselves.

        """
        if _is_dunder(name):
            raise AttributeError(name)
        try:
            return cls._member_map_[name]
        except KeyError:
            raise AttributeError(name)

    def __getitem__(cls, name):
        return cls._member_map_[name]

    def __iter__(cls):
        return (cls._member_map_[name] for name in cls._member_names_)

    def __reversed__(cls):
        return (cls._member_map_[name] for name in reversed(cls._member_names_))

    def __len__(cls):
        return len(cls._member_names_)

    def __repr__(cls):
        return "<enum %r>" % cls.__name__

    def __setattr__(cls, name, value):
        """Block attempts to reassign Enum members.

        A simple assignment to the class namespace only changes one of the
        several possible ways to get an Enum member from the Enum class,
        resulting in an inconsistent Enumeration.

        """
        member_map = cls.__dict__.get('_member_map_', {})
        if name in member_map:
            raise AttributeError('Cannot reassign members.')
        super(EnumMeta, cls).__setattr__(name, value)

    def _create_(cls, class_name, names=None, module=None, type=None):
        """Convenience method to create a new Enum class.

        `names` can be:

        * A string containing member names, separated either with spaces or
          commas.  Values are auto-numbered from 1.
        * An iterable of member names.  Values are auto-numbered from 1.
        * An iterable of (member name, value) pairs.
        * A mapping of member name -> value.

        """
        metacls = cls.__class__
        if type is None:
            bases = (cls, )
        else:
            bases = (type, cls)
        classdict = metacls.__prepare__(class_name, bases)
        __order__ = []

        # special processing needed for names?
        if isinstance(names, str):
            names = names.replace(',', ' ').split()
        if isinstance(names, (tuple, list)) and isinstance(names[0], str):
            names = [(e, i+1) for (i, e) in enumerate(names)]

        # Here, names is either an iterable of (name, value) or a mapping.
        for item in names:
            if isinstance(item, str):
                member_name, member_value = item, names[item]
            else:
                member_name, member_value = item
            classdict[member_name] = member_value
            __order__.append(member_name)
        # only set __order__ in classdict if name/value was not from a mapping
        if not isinstance(item, str):
            classdict['__order__'] = ' '.join(__order__)
        enum_class = metacls.__new__(metacls, class_name, bases, classdict)

        # TODO: replace the frame hack if a blessed way to know the calling
        # module is ever developed
        if module is None:
            try:
                module = _sys._getframe(2).f_globals['__name__']
            except (AttributeError, ValueError):
                pass
        if module is None:
            _make_class_unpicklable(enum_class)
        else:
            enum_class.__module__ = module

        return enum_class

    @staticmethod
    def _get_mixins_(bases):
        """Returns the type for creating enum members, and the first inherited
        enum class.

        bases: the tuple of bases that was given to __new__

        """
        if not bases or Enum is None:
            return object, Enum
        

        # double check that we are not subclassing a class with existing
        # enumeration members; while we're at it, see if any other data
        # type has been mixed in so we can use the correct __new__
        member_type = first_enum = None
        for base in bases:
            if  (base is not Enum and
                    issubclass(base, Enum) and
                    base._member_names_):
                raise TypeError("Cannot extend enumerations")
        # base is now the last base in bases
        if not issubclass(base, Enum):
            raise TypeError("new enumerations must be created as "
                    "`ClassName([mixin_type,] enum_type)`")

        # get correct mix-in type (either mix-in type of Enum subclass, or
        # first base if last base is Enum)
        if not issubclass(bases[0], Enum):
            member_type = bases[0]     # first data type
            first_enum = bases[-1]  # enum type
        else:
            for base in bases[0].__mro__:
                # most common: (IntEnum, int, Enum, object)
                # possible:    (<Enum 'AutoIntEnum'>, <Enum 'IntEnum'>,
                #               <class 'int'>, <Enum 'Enum'>,
                #               <class 'object'>)
                if issubclass(base, Enum):
                    if first_enum is None:
                        first_enum = base
                else:
                    if member_type is None:
                        member_type = base

        return member_type, first_enum

    if pyver < 3.0:
        @staticmethod
        def _find_new_(classdict, member_type, first_enum):
            """Returns the __new__ to be used for creating the enum members.

            classdict: the class dictionary given to __new__
            member_type: the data type whose __new__ will be used by default
            first_enum: enumeration to check for an overriding __new__

            """
            # now find the correct __new__, checking to see of one was defined
            # by the user; also check earlier enum classes in case a __new__ was
            # saved as __member_new__
            __new__ = classdict.get('__new__', None)
            if __new__:
                return None, True, True      # __new__, save_new, use_args

            N__new__ = getattr(None, '__new__')
            O__new__ = getattr(object, '__new__')
            if Enum is None:
                E__new__ = N__new__
            else:
                E__new__ = Enum.__dict__['__new__']
            # check all possibles for __member_new__ before falling back to
            # __new__
            for method in ('__member_new__', '__new__'):
                for possible in (member_type, first_enum):
                    try:
                        target = possible.__dict__[method]
                    except (AttributeError, KeyError):
                        target = getattr(possible, method, None)
                    if target not in [
                            None,
                            N__new__,
                            O__new__,
                            E__new__,
                            ]:
                        if method == '__member_new__':
                            classdict['__new__'] = target
                            return None, False, True
                        if isinstance(target, staticmethod):
                            target = target.__get__(member_type)
                        __new__ = target
                        break
                if __new__ is not None:
                    break
            else:
                __new__ = object.__new__

            # if a non-object.__new__ is used then whatever value/tuple was
            # assigned to the enum member name will be passed to __new__ and to the
            # new enum member's __init__
            if __new__ is object.__new__:
                use_args = False
            else:
                use_args = True

            return __new__, False, use_args
    else:
        @staticmethod
        def _find_new_(classdict, member_type, first_enum):
            """Returns the __new__ to be used for creating the enum members.

            classdict: the class dictionary given to __new__
            member_type: the data type whose __new__ will be used by default
            first_enum: enumeration to check for an overriding __new__

            """
            # now find the correct __new__, checking to see of one was defined
            # by the user; also check earlier enum classes in case a __new__ was
            # saved as __member_new__
            __new__ = classdict.get('__new__', None)

            # should __new__ be saved as __member_new__ later?
            save_new = __new__ is not None

            if __new__ is None:
                # check all possibles for __member_new__ before falling back to
                # __new__
                for method in ('__member_new__', '__new__'):
                    for possible in (member_type, first_enum):
                        target = getattr(possible, method, None)
                        if target not in (
                                None,
                                None.__new__,
                                object.__new__,
                                Enum.__new__,
                                ):
                            __new__ = target
                            break
                    if __new__ is not None:
                        break
                else:
                    __new__ = object.__new__

            # if a non-object.__new__ is used then whatever value/tuple was
            # assigned to the enum member name will be passed to __new__ and to the
            # new enum member's __init__
            if __new__ is object.__new__:
                use_args = False
            else:
                use_args = True

            return __new__, save_new, use_args


########################################################
# In order to support Python 2 and 3 with a single
# codebase we have to create the Enum methods separately
# and then use the `type(name, bases, dict)` method to
# create the class.
########################################################
temp_enum_dict = {}
temp_enum_dict['__doc__'] = "Generic enumeration.\n\n    Derive from this class to define new enumerations.\n\n"

def __new__(cls, value):
    # all enum instances are actually created during class construction
    # without calling this method; this method is called by the metaclass'
    # __call__ (i.e. Color(3) ), and by pickle
    if type(value) is cls:
        # For lookups like Color(Color.red)
        value = value.value
        #return value
    # by-value search for a matching enum member
    # see if it's in the reverse mapping (for hashable values)
    try:
        if value in cls._value2member_map_:
            return cls._value2member_map_[value]
    except TypeError:
        # not there, now do long search -- O(n) behavior
        for member in cls._member_map_.values():
            if member.value == value:
                return member
    raise ValueError("%s is not a valid %s" % (value, cls.__name__))
temp_enum_dict['__new__'] = __new__
del __new__

def __repr__(self):
    return "<%s.%s: %r>" % (
            self.__class__.__name__, self._name_, self._value_)
temp_enum_dict['__repr__'] = __repr__
del __repr__

def __str__(self):
    return "%s.%s" % (self.__class__.__name__, self._name_)
temp_enum_dict['__str__'] = __str__
del __str__

def __dir__(self):
    added_behavior = [m for m in self.__class__.__dict__ if m[0] != '_']
    return (['__class__', '__doc__', '__module__', 'name', 'value'] + added_behavior)
temp_enum_dict['__dir__'] = __dir__
del __dir__

def __format__(self, format_spec):
    # mixed-in Enums should use the mixed-in type's __format__, otherwise
    # we can get strange results with the Enum name showing up instead of
    # the value

    # pure Enum branch
    if self._member_type_ is object:
        cls = str
        val = str(self)
    # mix-in branch
    else:
        cls = self._member_type_
        val = self.value
    return cls.__format__(val, format_spec)
temp_enum_dict['__format__'] = __format__
del __format__


####################################
# Python's less than 2.6 use __cmp__

if pyver < 2.6:

    def __cmp__(self, other):
        if type(other) is self.__class__:
            if self is other:
                return 0
            return -1
        return NotImplemented
        raise TypeError("unorderable types: %s() and %s()" % (self.__class__.__name__, other.__class__.__name__))
    temp_enum_dict['__cmp__'] = __cmp__
    del __cmp__

else:

    def __le__(self, other):
        raise TypeError("unorderable types: %s() <= %s()" % (self.__class__.__name__, other.__class__.__name__))
    temp_enum_dict['__le__'] = __le__
    del __le__

    def __lt__(self, other):
        raise TypeError("unorderable types: %s() < %s()" % (self.__class__.__name__, other.__class__.__name__))
    temp_enum_dict['__lt__'] = __lt__
    del __lt__

    def __ge__(self, other):
        raise TypeError("unorderable types: %s() >= %s()" % (self.__class__.__name__, other.__class__.__name__))
    temp_enum_dict['__ge__'] = __ge__
    del __ge__

    def __gt__(self, other):
        raise TypeError("unorderable types: %s() > %s()" % (self.__class__.__name__, other.__class__.__name__))
    temp_enum_dict['__gt__'] = __gt__
    del __gt__
    

def __eq__(self, other):
    if type(other) is self.__class__:
        return self is other
    return NotImplemented
temp_enum_dict['__eq__'] = __eq__
del __eq__

def __ne__(self, other):
    if type(other) is self.__class__:
        return self is not other
    return NotImplemented
temp_enum_dict['__ne__'] = __ne__
del __ne__

def __getnewargs__(self):
    return (self._value_, )
temp_enum_dict['__getnewargs__'] = __getnewargs__
del __getnewargs__

def __hash__(self):
    return hash(self._name_)
temp_enum_dict['__hash__'] = __hash__
del __hash__

# _RouteClassAttributeToGetattr is used to provide access to the `name`
# and `value` properties of enum members while keeping some measure of
# protection from modification, while still allowing for an enumeration
# to have members named `name` and `value`.  This works because enumeration
# members are not set directly on the enum class -- __getattr__ is
# used to look them up.

@_RouteClassAttributeToGetattr
def name(self):
    return self._name_
temp_enum_dict['name'] = name
del name

@_RouteClassAttributeToGetattr
def value(self):
    return self._value_
temp_enum_dict['value'] = value
del value

Enum = EnumMeta('Enum', (object, ), temp_enum_dict)
del temp_enum_dict

# Enum has now been created
###########################

class IntEnum(int, Enum):
    """Enum where members are also (and must be) ints"""


def unique(enumeration):
    """Class decorator that ensures only unique members exist in an enumeration."""
    duplicates = []
    for name, member in enumeration.__members__.items():
        if name != member.name:
            duplicates.append((name, member.name))
    if duplicates:
        duplicate_names = ', '.join(
                ["%s -> %s" % (alias, name) for (alias, name) in duplicates]
                )
        raise ValueError('duplicate names found in %r: %s' %
                (enumeration, duplicate_names)
                )
    return enumeration





#
# Currently supported auth types:
#      NTLM auth using SMBNTLM implementation
#      Kerberos auth using SMBSPNEGO implementation (GSSAPI)
#
# Not supported currently:
#      NTLM via SPNEGO (GSSAPI)
#      Direct kerberos auth (not sure if this is even in the protocol)
#      NETLOGON
#
# TODO: netlogon implementation (needs to be implemented in connection.py as well!)
# TODO: GSS over NTLM (but I'm not sure if the protocol allows that)

class DCERPCAuth:
    def __init__(self):
        self.ntlm = None
        self.kerberos = None
        self.gssapi = None
        self.netlogon = None

    @staticmethod
    def from_smb_gssapi(gssapi):
        auth = DCERPCAuth()
        auth.gssapi = gssapi.get_copy()
        if 'MS KRB5 - Microsoft Kerberos 5' in gssapi.list_original_conexts():
            auth.kerberos = gssapi.get_original_context('MS KRB5 - Microsoft Kerberos 5')
        if 'NTLMSSP - Microsoft NTLM Security Support Provider' in gssapi.list_original_conexts():
            auth.ntlm = gssapi.get_original_context('NTLMSSP - Microsoft NTLM Security Support Provider')

        return auth





from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.commons.interfaces.blocking.file.protocol import *

class SMBBlockingFileMgr:
	"""
	This class implamanets a file object manager which can open/read/write files over the smbconnection
	The input/output queues used for communicating are process/thread/async safe!
	Therefore this creates a bridge between a non-async thread/process and the smb connection (file ops only!)
	"""
	def __init__(self, connection, in_q, out_q):
		self.connection = connection
		self.in_q = in_q
		self.out_q = out_q
		self.filehandle = {}
		self.curhandle = 0

	async def run(self):
		while True:
			try:
				cmd = await self.in_q.coro_get()
				if cmd.cmd_type == SMBFILECMD.OPEN:
					try:
						sf = SMBFile.from_remotepath(self.connection, cmd.path)
						await sf.open(self.connection, cmd.mode)
						self.filehandle[self.curhandle] = sf
						self.curhandle += 1
						res = SMBFileOpenReply(cmd.cmd_id, self.curhandle -1)
						await self.out_q.coro_put(res)
					except Exception as e:
						res = SMBFileError(cmd.cmd_id, str(e))
						await self.out_q.coro_put(res)

				elif cmd.cmd_type == SMBFILECMD.READ:
					try:
						sf = self.filehandle.get(cmd.handle)
						await sf.seek(cmd.position, 0)
						data = await sf.read(cmd.count)
						res = SMBFileReadReply(cmd.cmd_id, cmd.handle, data)
						await self.out_q.coro_put(res)
					except Exception as e:
						res = SMBFileError(cmd.cmd_id, str(e))
						await self.out_q.coro_put(res)

				elif cmd.cmd_type == SMBFILECMD.CLOSE:
					try:
						sf = self.filehandle.get(cmd.handle)
						await sf.close()
						del self.filehandle[cmd.handle]
						res = SMBFileCloseReply(cmd.cmd_id, cmd.handle)
						await self.out_q.coro_put(res)
					except Exception as e:
						res = SMBFileError(cmd.cmd_id, str(e))
						await self.out_q.coro_put(res)

				elif cmd.cmd_type == SMBFILECMD.WRITE:
					try:
						sf = self.filehandle.get(cmd.handle)
						await sf.seek(cmd.position, 0)
						count = await sf.write(cmd.data)
						del self.filehandle[cmd.handle]
						res = SMBFileWriteReply(cmd.cmd_id, cmd.handle, count)
						await self.out_q.coro_put(res)
					except Exception as e:
						res = SMBFileError(cmd.cmd_id, str(e))
						await self.out_q.coro_put(res)

				elif cmd.cmd_type == SMBFILECMD.TERMINATE:
					for handle in self.filehandle:
						sf = self.filehandle[handle]
						await sf.close()
						del self.filehandle[handle]
					return
			except Exception as e:
				res = SMBFileError(cmd.cmd_id, str(e))
				await self.out_q.coro_put(res)
				return



import enum

class SMBServiceStatus(enum.Enum):
	CONTINUE_PENDING = 'CONTINUE_PENDING'
	PAUSE_PENDING = 'PAUSE_PENDING'
	PAUSED = 'PAUSED'
	RUNNING = 'RUNNING'
	START_PENDING = 'START_PENDING'
	STOP_PENDING = 'STOP_PENDING'
	STOPPED = 'STOPPED'
	DISABLED = 'DISABLED'
	UNKNOWN = 'UNKNOWN'

class SMBService:
	def __init__(self,name = None, display_name = None, status = None):
		self.name = name
		self.display_name = display_name
		self.status = status

	def __str__(self):
		return '%s - %s - %s' % (self.name, self.display_name, self.status.value)



import asyncio
import traceback

from aiosmb.dcerpc.v5.transport.selector import DCERPCTransportSelector
from aiosmb.dcerpc.v5.structure import Structure,pack,unpack
from aiosmb.dcerpc.v5 import uuid
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin, generate, stringver_to_bin, bin_to_uuidtup
from aiosmb.dcerpc.v5.dtypes import UCHAR, ULONG, USHORT
from aiosmb.dcerpc.v5.ndr import NDRSTRUCT
from aiosmb.dcerpc.v5.rpcrt import *
from aiosmb.commons.utils.decorators import red, rr
from minikerberos.gssapi.gssapi import GSSAPIFlags

class DCERPC5Connection:
	def __init__(self, gssapi, target):
		self.target = target
		self.gssapi = gssapi

		self.transport = None

		self.auth_type = RPC_C_AUTHN_WINNT
		if self.gssapi.kerberos is not None:
			self.auth_type = RPC_C_AUTHN_GSS_NEGOTIATE
		
		self.auth_level = RPC_C_AUTHN_LEVEL_NONE
		self.ctx = 0
		self.callid = 1

		self.NDRSyntax   = uuidtup_to_bin(('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0'))
		self.NDR64Syntax = uuidtup_to_bin(('71710533-BEBA-4937-8319-B5DBEF9CCC36', '1.0'))
		self.transfer_syntax =  self.NDRSyntax

		self.transfer_syntax = uuidtup_to_bin(('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0'))
		self.__max_xmit_size = 0
		self._max_user_frag = 0

		self.__sequence = 0
		self.__sessionKey = None
		self.__clientSigningKey = None
		self.__serverSigningKey = None
		self.__clientSealingKey = None
		self.__serverSealingKey = None
		self.__clientSealingHandle = None
		self.__serverSealingHandle = None

	def get_session_key(self):
		return self.__sessionKey

	def set_auth_level(self, auth_level):
		self.auth_level = auth_level

	def set_auth_type(self, auth_type):
		self.auth_type = auth_type

	@red
	async def disconnect(self):
		if self.transport is not None:
			await self.transport.disconnect()

		return True, None

	@red
	async def connect(self):
		"""
		Selects the correct transport layer based on the self.target and starts it
		"""
		selector = DCERPCTransportSelector()
		self.transport = await selector.select(self.target)
		await rr(self.transport.connect())
		
		return True, None

	@red
	async def bind(self, iface_uuid, alter = 0, bogus_binds = 0, transfer_syntax = ('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0')):
		"""
		Performs bind operation. Does authentication and sets up the keys for further communication
		"""
		bind = MSRPCBind()
		#item['TransferSyntax']['Version'] = 1
		ctx = self.ctx
		for _ in range(bogus_binds):
			item = CtxItem()
			item['ContextID'] = ctx
			item['TransItems'] = 1
			item['ContextID'] = ctx
			# We generate random UUIDs for bogus binds
			item['AbstractSyntax'] = generate() + stringver_to_bin('2.0')
			item['TransferSyntax'] = uuidtup_to_bin(transfer_syntax)
			bind.addCtxItem(item)
			self.ctx += 1
			ctx += 1

		# The true one :)
		item = CtxItem()
		item['AbstractSyntax'] = iface_uuid
		item['TransferSyntax'] = uuidtup_to_bin(transfer_syntax)
		item['ContextID'] = ctx
		item['TransItems'] = 1
		bind.addCtxItem(item)

		packet = MSRPCHeader()
		packet['type'] = MSRPC_BIND
		packet['pduData'] = bind.getData()
		packet['call_id'] = self.callid
		
		if alter:
			packet['type'] = MSRPC_ALTERCTX

		if self.auth_level != RPC_C_AUTHN_LEVEL_NONE:
			#authentication required
			if self.auth_type == RPC_C_AUTHN_WINNT:
				
				#seal flag MUST be turned on in the handshake flags!!!!!!!
				#it is "signaled via the is_rpc variable"
				auth, res = await self.gssapi.ntlm.authenticate(None, is_rpc = True)

			elif self.auth_type == RPC_C_AUTHN_NETLOGON:
				raise Exception('Not implemented!')

			elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
				auth, res  = await self.gssapi.gssapi.authenticate(
					None, 
					flags = GSSAPIFlags.GSS_C_CONF_FLAG |\
							GSSAPIFlags.GSS_C_INTEG_FLAG | \
							GSSAPIFlags.GSS_C_SEQUENCE_FLAG | \
							GSSAPIFlags.GSS_C_REPLAY_FLAG | \
							GSSAPIFlags.GSS_C_MUTUAL_FLAG | \
							GSSAPIFlags.GSS_C_DCE_STYLE,
					seq_number = 0, 
					is_rpc = True
				)
			else:
				Exception('Unsupported auth type!')

			sec_trailer = SEC_TRAILER()
			sec_trailer['auth_type']   = self.auth_type
			sec_trailer['auth_level']  = self.auth_level
			sec_trailer['auth_ctx_id'] = self.ctx + 79231 

			pad = (4 - (len(packet.get_packet()) % 4)) % 4
			if pad != 0:
				packet['pduData'] += b'\xFF'*pad
				sec_trailer['auth_pad_len']=pad

			packet['sec_trailer'] = sec_trailer
			packet['auth_data'] = auth

		_,_ = await rr(self.transport.send(packet.get_packet()))
		
		data, err = await rr(self.recv_one())

		resp = MSRPCHeader(data)

		if resp['type'] == MSRPC_BINDACK or resp['type'] == MSRPC_ALTERCTX_R:
			bindResp = MSRPCBindAck(resp.getData())
		elif resp['type'] == MSRPC_BINDNAK or resp['type'] == MSRPC_FAULT:
			if resp['type'] == MSRPC_FAULT:
				resp = MSRPCRespHeader(resp.getData())
				status_code = unpack('<L', resp['pduData'][:4])[0]
			else:
				resp = MSRPCBindNak(resp['pduData'])
				status_code = resp['RejectedReason']
			if status_code in rpc_status_codes:
				raise DCERPCException(error_code = status_code)
			elif status_code in rpc_provider_reason:
				raise DCERPCException("Bind context rejected: %s" % rpc_provider_reason[status_code])
			else:
				raise DCERPCException('Unknown DCE RPC fault status code: %.8x' % status_code)
		else:
			raise DCERPCException('Unknown DCE RPC packet type received: %d' % resp['type'])

		# check ack results for each context, except for the bogus ones
		for ctx in range(bogus_binds+1,bindResp['ctx_num']+1):
			ctxItems = bindResp.getCtxItem(ctx)
			if ctxItems['Result'] != 0:
				msg = "Bind context %d rejected: " % ctx
				msg += rpc_cont_def_result.get(ctxItems['Result'], 'Unknown DCE RPC context result code: %.4x' % ctxItems['Result'])
				msg += "; "
				reason = bindResp.getCtxItem(ctx)['Reason']
				msg += rpc_provider_reason.get(reason, 'Unknown reason code: %.4x' % reason)
				if (ctxItems['Result'], reason) == (2, 1): # provider_rejection, abstract syntax not supported
					msg += " (this usually means the interface isn't listening on the given endpoint)"
				raise DCERPCException(msg)

			# Save the transfer syntax for later use
			self.transfer_syntax = ctxItems['TransferSyntax']

		# The received transmit size becomes the client's receive size, and the received receive size becomes the client's transmit size.
		self.__max_xmit_size = bindResp['max_rfrag']

		if self.auth_level != RPC_C_AUTHN_LEVEL_NONE:
			if self.auth_type == RPC_C_AUTHN_WINNT:
				response, res = await self.gssapi.ntlm.authenticate(bindResp['auth_data'], is_rpc = True)
				self.__sessionKey = self.gssapi.ntlm.get_session_key()
				

			elif self.auth_type == RPC_C_AUTHN_NETLOGON:
				response = None
			elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
				response, res  = await self.gssapi.gssapi.authenticate(
					bindResp['auth_data'], 
					is_rpc = True, 
					flags = GSSAPIFlags.GSS_C_CONF_FLAG |\
						GSSAPIFlags.GSS_C_INTEG_FLAG | \
						GSSAPIFlags.GSS_C_SEQUENCE_FLAG | \
						GSSAPIFlags.GSS_C_REPLAY_FLAG | \
						GSSAPIFlags.GSS_C_MUTUAL_FLAG | \
						GSSAPIFlags.GSS_C_DCE_STYLE
				)
																							
				self.__sessionKey = self.gssapi.gssapi.get_session_key()

			self.__sequence = 0

			if self.auth_level in (RPC_C_AUTHN_LEVEL_CONNECT, RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY):
				if self.auth_type == RPC_C_AUTHN_WINNT:
					if self.gssapi.ntlm.is_extended_security() == True:
						self.__clientSigningKey = self.gssapi.ntlm.get_signkey() 
						self.__serverSigningKey = self.gssapi.ntlm.get_signkey('Server')
						self.__clientSealingKey = self.gssapi.ntlm.get_sealkey() 
						self.__serverSealingKey = self.gssapi.ntlm.get_sealkey('Server')
						cipher3 = RC4(self.__clientSealingKey)
						self.__clientSealingHandle = cipher3.encrypt
						cipher4 = RC4(self.__serverSealingKey)
						self.__serverSealingHandle = cipher4.encrypt
						
					else:
						# Same key for everything
						self.__clientSigningKey = self.gssapi.ntlm.get_session_key()
						self.__serverSigningKey = self.gssapi.ntlm.get_session_key()
						self.__clientSealingKey = self.gssapi.ntlm.get_session_key()
						self.__serverSealingKey = self.gssapi.ntlm.get_session_key()
						cipher = RC4(self.__clientSigningKey)
						self.__clientSealingHandle = cipher.encrypt
						self.__serverSealingHandle = cipher.encrypt
					
				elif self.auth_type == RPC_C_AUTHN_NETLOGON:
					raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')

			sec_trailer = SEC_TRAILER()
			sec_trailer['auth_type'] = self.auth_type
			sec_trailer['auth_level'] = self.auth_level
			sec_trailer['auth_ctx_id'] = self.ctx + 79231 

			if response is not None:
				if self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
					alter_ctx = MSRPCHeader()
					alter_ctx['type'] = MSRPC_ALTERCTX
					alter_ctx['pduData'] = bind.getData()
					alter_ctx['sec_trailer'] = sec_trailer
					alter_ctx['auth_data'] = response
					
					await rr(self.transport.send(alter_ctx.get_packet(), forceWriteAndx = 1))
					
					self.__sequence = 0
					await rr(self.recv_one()) #recieving the result of alter_context command

					self.__sequence = 0
				else:
					auth3 = MSRPCHeader()
					auth3['type'] = MSRPC_AUTH3
					# pad (4 bytes): Can be set to any arbitrary value when set and MUST be 
					# ignored on receipt. The pad field MUST be immediately followed by a 
					# sec_trailer structure whose layout, location, and alignment are as 
					# specified in section 2.2.2.11
					auth3['pduData'] = b' ' * 4 #SkelSec: I have spent 3 hours to find this bug, that I caused by replacing spaces to tabs :(
					auth3['sec_trailer'] = sec_trailer
					#SkelSec auth3['auth_data'] = response.getData()
					auth3['auth_data'] = response

					# Use the same call_id
					self.callid = resp['call_id']
					auth3['call_id'] = self.callid
					await rr(self.transport.send(auth3.get_packet(), forceWriteAndx = 1))

			self.callid += 1

		return resp, None	 # means packet is signed, if verifier is wrong it fails

	@red
	async def request(self, request, uuid=None, checkError=True):
		"""
		Creates a requests then dispateches it to _transport.send for singing/encryption asn sending
		"""

		if self.transfer_syntax == self.NDR64Syntax:
			request.changeTransferSyntax(self.NDR64Syntax)
			isNDR64 = True
		else:
			isNDR64 = False
			
		await rr(self.call(request.opnum, request, uuid))
		answer, _ = await rr(self.recv())

		__import__(request.__module__)
		module = sys.modules[request.__module__]
		respClass = getattr(module, request.__class__.__name__ + 'Response')

		if answer[-4:] != b'\x00\x00\x00\x00' and checkError is True:
			error_code = unpack('<L', answer[-4:])[0]
			if error_code in rpc_status_codes:
				
				# This is an error we can handle
				exception = DCERPCException(error_code = error_code)
			else:	
				
				sessionErrorClass = getattr(module, 'DCERPCSessionError')
				try:
					# Try to unpack the answer, even if it is an error, it works most of the times
					response =  respClass(answer, isNDR64 = isNDR64)
				except:
					# No luck :(
					exception = sessionErrorClass(error_code = error_code)
				else:
					exception = sessionErrorClass(packet = response, error_code = error_code)
			return None, exception
		else:
			response =  respClass(answer, isNDR64 = isNDR64)
			return response, None

	@red
	async def send(self, data):
	
		if isinstance(data, MSRPCHeader) is not True:
			# Must be an Impacket, transform to structure
			data = DCERPC_RawCall(data.OP_NUM, data.get_packet())

		try:
			if data['uuid'] != b'':
				data['flags'] |= PFC_OBJECT_UUID
		except:
			# Structure doesn't have uuid
			pass
		data['ctx_id'] = self.ctx
		data['call_id'] = self.callid
		data['alloc_hint'] = len(data['pduData'])
		# We should fragment PDUs if:
		# 1) Payload exceeds __max_xmit_size received during BIND response
		# 2) We'e explicitly fragmenting packets with lower values
		should_fragment = False

		# Let's decide what will drive fragmentation for this request
		if self._max_user_frag > 0:
			# User set a frag size, let's compare it with the max transmit size agreed when binding the interface
			fragment_size = min(self._max_user_frag, self.__max_xmit_size)
		else:
			fragment_size = self.__max_xmit_size

		# Sanity check. Fragmentation can't be too low, otherwise sec_trailer won't fit

		if self.auth_level in [RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY]:
			if fragment_size <= 8:
				# Minimum pdu fragment size is 8, important when doing PKT_INTEGRITY/PRIVACY. We need a minimum size of 8
				# (Kerberos)
				fragment_size = 8

		# ToDo: Better calculate the size needed. Now I'm setting a number that surely is enough for Kerberos and NTLM
		# ToDo: trailers, both for INTEGRITY and PRIVACY. This means we're not truly honoring the user's frag request.
		if len(data['pduData']) + 128 > fragment_size:
			should_fragment = True
			if fragment_size+128 > self.__max_xmit_size:
				fragment_size = self.__max_xmit_size - 128

		if should_fragment:
			packet = data['pduData']
			offset = 0

			while 1:
				toSend = packet[offset:offset+fragment_size]
				if not toSend:
					break
				if offset == 0:
					data['flags'] |= PFC_FIRST_FRAG
				else:
					data['flags'] &= (~PFC_FIRST_FRAG)
				offset += len(toSend)
				if offset >= len(packet):
					data['flags'] |= PFC_LAST_FRAG
				else:
					data['flags'] &= (~PFC_LAST_FRAG)
				data['pduData'] = toSend
				self._transport_send(data, forceWriteAndx = 1, forceRecv =data['flags'] & PFC_LAST_FRAG)
		else:
			await rr(self._transport_send(data))
		self.callid += 1
		return True, None

	@red
	async def recv(self):

		finished = False
		forceRecv = 0
		retAnswer = b''
		while not finished:
			# At least give me the MSRPCRespHeader, especially important for 
			# TCP/UDP Transports
			response_data, _ = await rr(self.transport.recv(MSRPCRespHeader._SIZE))
			
			response_header = MSRPCRespHeader(response_data)
			# Ok, there might be situation, especially with large packets, that 
			# the transport layer didn't send us the full packet's contents
			# So we gotta check we received it all
			while len(response_data) < response_header['frag_len']:
				data, _ = await rr(self.transport.recv(response_header['frag_len']-len(response_data)))
				response_data += data

			off = response_header.get_header_size()
			
			if response_header['type'] == MSRPC_FAULT and response_header['frag_len'] >= off+4:
				status_code = unpack("<L",response_data[off:off+4])[0]
				if status_code in rpc_status_codes:
					raise DCERPCException(rpc_status_codes[status_code])
				elif status_code & 0xffff in rpc_status_codes:
					raise DCERPCException(rpc_status_codes[status_code & 0xffff])
				else:
					if status_code in hresult_errors.ERROR_MESSAGES:
						error_msg_short = hresult_errors.ERROR_MESSAGES[status_code][0]
						error_msg_verbose = hresult_errors.ERROR_MESSAGES[status_code][1] 
						raise DCERPCException('%s - %s' % (error_msg_short, error_msg_verbose))
					else:
						raise DCERPCException('Unknown DCE RPC fault status code: %.8x' % status_code)

			if response_header['flags'] & PFC_LAST_FRAG:
				# No need to reassembly DCERPC
				finished = True
			else:
				# Forcing Read Recv, we need more packets!
				forceRecv = 1
			
			#print('response_data: %s' % response_data.hex())
			answer = response_data[off:]
			auth_len = response_header['auth_len']
			if auth_len:
				auth_len += 8
				auth_data = answer[-auth_len:]
				sec_trailer = SEC_TRAILER(data = auth_data)
				answer = answer[:-auth_len]
				
				if sec_trailer['auth_level'] == RPC_C_AUTHN_LEVEL_PKT_PRIVACY:
					if self.auth_type == RPC_C_AUTHN_WINNT:
						if self.gssapi.ntlm.is_extended_security() == True:
							# TODO: FIX THIS, it's not calculating the signature well
							# Since I'm not testing it we don't care... yet
							answer, signature = self.gssapi.ntlm.SEAL(
														self.__serverSigningKey,
														self.__serverSealingKey,
														answer,
														answer,
														self.__sequence,
														self.__serverSealingHandle
													)

						else:
							answer, signature = self.gssapi.ntlm.SEAL(
														self.__serverSigningKey,
														self.__serverSealingKey,
														answer,
														answer,
														self.__sequence,
														self.__serverSealingHandle
													)

							self.__sequence += 1
					elif self.auth_type == RPC_C_AUTHN_NETLOGON:
						raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
						#from impacket.dcerpc.v5 import nrpc
						#answer, cfounder = nrpc.UNSEAL(answer, 
						#	   auth_data[len(sec_trailer):],
						#	   self.__sessionKey, 
						#	   False)
						#self.__sequence += 1
					elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
						if self.__sequence > 0:
							answer, cfounder = await self.gssapi.gssapi.decrypt(answer, self.__sequence, direction='init', auth_data=auth_data)
																	

				elif sec_trailer['auth_level'] == RPC_C_AUTHN_LEVEL_PKT_INTEGRITY:
					if self.auth_type == RPC_C_AUTHN_WINNT:
						ntlmssp = auth_data[12:]
						
						if self.gssapi.ntlm.is_extended_security() == True:
							#TODO:
							signature =  self.gssapi.ntlm.SIGN(
								self.__serverSigningKey, 
								answer, 
								self.__sequence, 
								self.__serverSealingHandle
							)
						else:
							signature = self.gssapi.ntlm.SIGN(
								self.__serverSigningKey, 
								ntlmssp, 
								self.__sequence, 
								self.__serverSealingHandle
							)
							# Yes.. NTLM2 doesn't increment sequence when receiving
							# the packet :P
							self.__sequence += 1
					elif self.auth_type == RPC_C_AUTHN_NETLOGON:
						raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
						#from impacket.dcerpc.v5 import nrpc
						#ntlmssp = auth_data[12:]
						#signature = nrpc.SIGN(ntlmssp, 
						#	   self.__confounder, 
						#	   self.__sequence, 
						#	   self.__sessionKey, 
						#	   False)
						#self.__sequence += 1
					elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
						# Do NOT increment the sequence number when Signing Kerberos
						#self.__sequence += 1
						pass

				
				if sec_trailer['auth_pad_len']:
					answer = answer[:-sec_trailer['auth_pad_len']]
			
			retAnswer += answer
		
		return retAnswer, None

	@red
	async def recv_one(self):
		finished = False
		forceRecv = 0
		retAnswer = b''
		while not finished:
			# At least give me the MSRPCRespHeader, especially important for 
			# TCP/UDP Transports
			
			response_data, _ = await rr(self.transport.recv(MSRPCRespHeader._SIZE))
			#print('DATA: %s' % repr(response_data))
			response_header = MSRPCRespHeader(response_data)
			# Ok, there might be situation, especially with large packets, that 
			# the transport layer didn't send us the full packet's contents
			# So we gotta check we received it all
			while len(response_data) < response_header['frag_len']:
				data, _ = await rr(self.transport.recv(response_header['frag_len']-len(response_data)))
				#print('DATA1: %s' % repr(response_data))
				response_data += data

			off = response_header.get_header_size()

			if response_header['type'] == MSRPC_FAULT and response_header['frag_len'] >= off+4:
				status_code = unpack("<L",response_data[off:off+4])[0]
				if status_code in rpc_status_codes:
					return None, DCERPCException(rpc_status_codes[status_code])
				elif status_code & 0xffff in rpc_status_codes:
					return None,  DCERPCException(rpc_status_codes[status_code & 0xffff])
				else:
					if status_code in hresult_errors.ERROR_MESSAGES:
						error_msg_short = hresult_errors.ERROR_MESSAGES[status_code][0]
						error_msg_verbose = hresult_errors.ERROR_MESSAGES[status_code][1] 
						return None,  DCERPCException('%s - %s' % (error_msg_short, error_msg_verbose))
					else:
						return None,  DCERPCException('Unknown DCE RPC fault status code: %.8x' % status_code)

			if response_header['flags'] & PFC_LAST_FRAG:
				# No need to reassembly DCERPC
				finished = True
			else:
				# Forcing Read Recv, we need more packets!
				forceRecv = 1
				
		return response_data, None
	

	@red
	async def _transport_send(self, rpc_packet, forceWriteAndx = 0, forceRecv = 0):
		"""
		This function does the signing and ecryption on the data we want to send out.
		Keys are set up during bind
		"""

		rpc_packet['ctx_id'] = self.ctx
		rpc_packet['sec_trailer'] = b''
		rpc_packet['auth_data'] = b''

		if self.auth_level in [RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY]:
			# Dummy verifier, just for the calculations
			sec_trailer = SEC_TRAILER()
			sec_trailer['auth_type'] = self.auth_type
			sec_trailer['auth_level'] = self.auth_level
			sec_trailer['auth_pad_len'] = 0
			sec_trailer['auth_ctx_id'] = self.ctx + 79231 

			pad = (4 - (len(rpc_packet.get_packet()) % 4)) % 4
			if pad != 0:
				rpc_packet['pduData'] += b'\xBB'*pad
				sec_trailer['auth_pad_len']=pad

			rpc_packet['sec_trailer'] = sec_trailer.getData()
			rpc_packet['auth_data'] = b' '*16

			plain_data = rpc_packet['pduData']
			if self.auth_level == RPC_C_AUTHN_LEVEL_PKT_PRIVACY:
				if self.auth_type == RPC_C_AUTHN_WINNT:
					if self.gssapi.ntlm.is_extended_security() == True:
						sealedMessage, signature = self.gssapi.ntlm.SEAL(
							self.__clientSigningKey, 
							self.__clientSealingKey,
							rpc_packet.get_packet()[:-16],
							plain_data,
							self.__sequence,
							self.__clientSealingHandle
						)
					else:
						sealedMessage, signature = self.gssapi.ntlm.SEAL(
							self.__clientSigningKey, 
							self.__clientSealingKey,
							plain_data,
							plain_data,
							self.__sequence,
							self.__clientSealingHandle
						)
					
				elif self.auth_type == RPC_C_AUTHN_NETLOGON:
					raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
					#from impacket.dcerpc.v5 import nrpc
					#sealedMessage, signature = nrpc.SEAL(plain_data, self.__confounder, self.__sequence, self.__sessionKey, False)
				elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
					#sealedMessage, signature = self.__gss.GSS_Wrap(self.__sessionKey, plain_data, self.__sequence)
					#print('HERE!')
					sealedMessage, signature = await self.gssapi.gssapi.encrypt(plain_data, self.__sequence)
					#sealedMessage, signature = await self.gssapi.encrypt(plain_data, self.__sequence)


				rpc_packet['pduData'] = sealedMessage
				#print(sealedMessage)
				#print(self.__auth_level)
			elif self.auth_level == RPC_C_AUTHN_LEVEL_PKT_INTEGRITY: 
				if self.auth_type == RPC_C_AUTHN_WINNT:
					if self.gssapi.ntlm.is_extended_security() == True:
						# Interesting thing.. with NTLM2, what is is signed is the 
						# whole PDU, not just the data
						signature =  self.gssapi.ntlm.SIGN(self.__clientSigningKey, 
							rpc_packet.get_packet()[:-16], 
							self.__sequence, 
							self.__clientSealingHandle)
					else:
						signature =  self.gssapi.ntlm.SIGN(self.__clientSigningKey, 
							plain_data, 
							self.__sequence, 
							self.__clientSealingHandle)
				elif self.auth_type == RPC_C_AUTHN_NETLOGON:
					raise Exception('RPC_C_AUTHN_NETLOGON is not implemented!')
					#from impacket.dcerpc.v5 import nrpc
					#signature = nrpc.SIGN(plain_data, 
					#	   self.__confounder, 
					#	   self.__sequence, 
					#	   self.__sessionKey, 
					#	   False)
				elif self.auth_type == RPC_C_AUTHN_GSS_NEGOTIATE:
					#signature = self.__gss.GSS_GetMIC(self.__sessionKey, plain_data, self.__sequence)
					signature = await self.gssapi.gssapi.sign(plain_data, self.__sequence)

			rpc_packet['sec_trailer'] = sec_trailer.getData()
			rpc_packet['auth_data'] = signature

			self.__sequence += 1

		await rr(self.transport.send(rpc_packet.get_packet(), forceWriteAndx = forceWriteAndx, forceRecv = forceRecv))
		return True, None

	@red
	async def call(self, function, body, uuid=None):
		if hasattr(body, 'getData'):
			t, _ = await rr(self.send(DCERPC_RawCall(function, body.getData(), uuid)))
			return t, None
		else:
			t, _ = await rr(self.send(DCERPC_RawCall(function, body, uuid)))
			return t, None

		
	def alter_ctx(self, newUID, bogus_binds = 0):
		answer = self.__class__(self.transport)

		answer.set_credentials(self.__username, self.__password, self.__domain, self.__lmhash, self.__nthash,
							   self.__aesKey, self.__TGT, self.__TGS)
		answer.set_auth_type(self.__auth_type)
		answer.set_auth_level(self.__auth_level)

		answer.set_ctx_id(self.ctx+1)
		answer.__callid = self.callid
		answer.bind(newUID, alter = 1, bogus_binds = bogus_binds, transfer_syntax = bin_to_uuidtup(self.transfer_syntax))
		return answer


import io
import enum
import sys

from aiosmb.protocol.smb.commands import *
from aiosmb.protocol.smb.header import *

# https://msdn.microsoft.com/en-us/library/ee441774.aspx
class SMBMessage:
	"""
	Base class for all SMB version 1 messages
	"""
	def __init__(self, header = None, command = None):
		self.header  = header #https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cifs/69a29f73-de0c-45a6-a1aa-8ceeea42217f
		self.command = command

	@staticmethod
	def from_bytes(bbuff):
		return SMBMessage.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SMBMessage()
		msg.header = SMBHeader.from_buffer(buff)
		classname = msg.header.Command.name
		if SMBHeaderFlagsEnum.SMB_FLAGS_REPLY in msg.header.Flags:
			classname += '_REPLY'
		else:
			classname += '_REQ'
		class_ = getattr(sys.modules[__name__], classname)
		msg.command = class_.from_buffer(buff)
		
		return msg

	def to_bytes(self):
		t  = self.header.to_bytes()
		t += self.command.to_bytes() 
		return t

	def __repr__(self):
		t = '== SMBMessage ==\r\n'
		t += repr(self.header)
		t += repr(self.command)
		return t





from aiosmb.authentication.ntlm.structures.avpair import AVPAIRType
from aiosmb.wintypes.dtyp.structures.filetime import FILETIME
import datetime

class NTLMServerInfo:
	def __init__(self):
		self.domainname = None
		self.computername = None
		self.dnscomputername = None
		self.dnsdomainname = None
		self.local_time = None
		self.dnsforestname = None
		self.os_major_version = None
		self.os_minor_version = None
		self.os_build = None
		self.os_guess = None
	
	@staticmethod
	def from_challenge(challenge):
		si = NTLMServerInfo()
		ti = challenge.TargetInfo
		for k in ti:
			if k == AVPAIRType.MsvAvNbDomainName:
				si.domainname = ti[k]
			elif k == AVPAIRType.MsvAvNbComputerName:
				si.computername = ti[k]
			elif k == AVPAIRType.MsvAvDnsDomainName:
				si.dnsdomainname = ti[k]
			elif k == AVPAIRType.MsvAvDnsComputerName:
				si.dnscomputername = ti[k]
			elif k == AVPAIRType.MsvAvDnsTreeName:
				si.dnsforestname = ti[k]
			elif k == AVPAIRType.MsvAvTimestamp:
				if isinstance(ti[k], bytes):
					si.local_time = FILETIME.from_bytes(ti[k]).datetime
				elif isinstance(ti[k], datetime):
					si.local_time = ti[k]
		
		if challenge.Version is not None:
			si.os_major_version = challenge.Version.ProductMajorVersion
			si.os_minor_version = challenge.Version.ProductMinorVersion
			si.os_build = challenge.Version.ProductBuild
			si.os_guess = challenge.Version.WindowsProduct
				
		return si

	def to_dict(self):
		return {
			'domainname' : self.domainname,
			'computername' : self.computername,
			'dnscomputername' : self.dnscomputername,
			'dnsdomainname' : self.dnsdomainname,
			'local_time' : self.local_time,
			'dnsforestname' : self.dnsforestname,
			'os_major_version' : self.os_major_version.name if self.os_major_version.name is not None else None,
			'os_minor_version' : self.os_minor_version.name if self.os_minor_version.name is not None else None,
			'os_build' : self.os_build,
			'os_guess' : self.os_guess,
		}
		
	def __str__(self):
		t = '=== Server Info ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k]) 
			
		return t


import enum

# https://msdn.microsoft.com/en-us/library/cc246528.aspx
class SMB2Command(enum.Enum):
	NEGOTIATE       = 0x0000
	SESSION_SETUP   = 0x0001
	LOGOFF          = 0x0002
	TREE_CONNECT    = 0x0003
	TREE_DISCONNECT = 0x0004
	CREATE          = 0x0005
	CLOSE           = 0x0006
	FLUSH           = 0x0007
	READ            = 0x0008
	WRITE           = 0x0009
	LOCK            = 0x000A
	IOCTL           = 0x000B
	CANCEL          = 0x000C
	ECHO            = 0x000D
	QUERY_DIRECTORY = 0x000E
	CHANGE_NOTIFY   = 0x000F
	QUERY_INFO      = 0x0010
	SET_INFO        = 0x0011
	OPLOCK_BREAK    = 0x0012


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# Comments:
#


import ipaddress
import enum
import copy

from aiosmb.protocol.common import NegotiateDialects, SMB2_NEGOTIATE_DIALTECTS_2, SMB2_NEGOTIATE_DIALTECTS_3, SMB2_NEGOTIATE_DIALTECTS

class SMBConnectionDialect(enum.Enum):
	SMB = 'SMB' #any, will us a wildcard because SMB1 is not implremented
	SMB2 = 'SMB2' #will offer all v2 versions
	SMB3 = 'SMB3' #will offer all v3 versions
	SMB202 = 'SMB202'
	SMB210 = 'SMB210'
	#SMB222 = 'SMB222'
	#SMB224 = 'SMB224'
	SMB300 = 'SMB300'
	SMB302 = 'SMB302'
	#SMB310 = 'SMB310'
	SMB311 = 'SMB311'

smb_negotiate_dialect_lookup = {
	SMBConnectionDialect.SMB202 : NegotiateDialects.SMB202,
	SMBConnectionDialect.SMB210 : NegotiateDialects.SMB210,
	#SMBConnectionDialect.SMB222 : NegotiateDialects.SMB222,
	#SMBConnectionDialect.SMB224 : NegotiateDialects.SMB224,
	SMBConnectionDialect.SMB300 : NegotiateDialects.SMB300,
	SMBConnectionDialect.SMB302 : NegotiateDialects.SMB302,
	#SMBConnectionDialect.SMB310 : NegotiateDialects.SMB310,
	SMBConnectionDialect.SMB311 : NegotiateDialects.SMB311,
}

class SMBConnectionProtocol(enum.Enum):
	TCP = 'TCP'
	UDP = 'UDP'

class SMBTarget:
	"""
	"""
	def __init__(self, ip = None, 
						port = 445, 
						hostname = None, 
						timeout = 1, 
						dc_ip=None, 
						domain = None, 
						proxy = None,
						protocol = SMBConnectionProtocol.TCP):
		self.ip = ip
		self.port = port
		self.hostname = hostname
		self.timeout = timeout
		self.dc_ip = dc_ip
		self.domain = domain
		self.proxy = proxy
		self.protocol = protocol
		self.preferred_dialects = SMB2_NEGOTIATE_DIALTECTS_2


	def update_dialect(self, dialect):
		if isinstance(dialect, SMBConnectionDialect) is False:
			raise Exception('dialect must be a type of SMBConnectionDialect')
		if dialect == SMBConnectionDialect.SMB:
			self.preferred_dialects = SMB2_NEGOTIATE_DIALTECTS
			self.preferred_dialects[NegotiateDialects.WILDCARD] = 1
		elif dialect == SMBConnectionDialect.SMB2:
			self.preferred_dialects = SMB2_NEGOTIATE_DIALTECTS_2
		elif dialect == SMBConnectionDialect.SMB3:
			self.preferred_dialects = SMB2_NEGOTIATE_DIALTECTS_3
			
		else:
			self.preferred_dialects = {
				smb_negotiate_dialect_lookup[dialect] : 1,
				NegotiateDialects.WILDCARD : 1,
			}
		return

	def to_target_string(self):
		return 'cifs/%s@%s' % (self.hostname, self.domain)

	def get_copy(self, ip, port, hostname = None):
		t = SMBTarget(
			ip = ip, 
			port = port, 
			hostname = hostname, 
			timeout = self.timeout, 
			dc_ip= self.dc_ip, 
			domain = self.domain, 
			proxy = copy.deepcopy(self.proxy),
			protocol = self.protocol
		)

		self.preferred_dialects
		return t
	
	@staticmethod
	def from_connection_string(s):
		port = 445
		dc = None
		
		_, target = s.rsplit('@', 1)
		if target.find('/') != -1:
			target, dc = target.split('/')
			
		if target.find(':') != -1:
			target, port = target.split(':')
			
		st = SMBTarget()
		st.port = port
		st.dc_ip = dc
		st.domain, _ = s.split('/', 1)
		
		try:
			st.ip = str(ipaddress.ip_address(target))
		except:
			st.hostname = target
	
		return st
		
	def get_ip(self):
		if not self.ip and not self.hostname:
			raise Exception('SMBTarget must have ip or hostname defined!')
		return self.ip if self.ip is not None else self.hostname
		
	def get_hostname(self):
		return self.hostname
	
	def get_hostname_or_ip(self):
		if self.hostname:
			return self.hostname
		return self.ip
	
	def get_port(self):
		return self.port
		
	def __str__(self):
		t = '==== SMBTarget ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t
		
		
def test():
	s = 'TEST/victim/ntlm/nt:AAAAAAAA@10.10.10.2:445'
	creds = SMBTarget.from_connection_string(s)
	print(str(creds))
	
	s = 'TEST/victim/sspi@10.10.10.2:445/aaaa'
	creds = SMBTarget.from_connection_string(s)
	
	print(str(creds))
	
if __name__ == '__main__':
	test()


import io
import enum

from aiosmb.wintypes.dtyp.constrcuted_security.acl import ACL
from aiosmb.wintypes.dtyp.constrcuted_security.sid import SID

class SE_SACL(enum.IntFlag):
	SE_DACL_AUTO_INHERIT_REQ = 0x0100 	#Indicates a required security descriptor in which the discretionary access control list (DACL) is set up to support automatic propagation of inheritable access control entries (ACEs) to existing child objects.
										#For access control lists (ACLs) that support auto inheritance, this bit is always set. Protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function to convert a security descriptor and set this flag.
	SE_DACL_AUTO_INHERITED = 0x0400     #Indicates a security descriptor in which the discretionary access control list (DACL) is set up to support automatic propagation of inheritable access control entries (ACEs) to existing child objects.
										#For access control lists (ACLs) that support auto inheritance, this bit is always set. Protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function to convert a security descriptor and set this flag.
	SE_DACL_DEFAULTED = 0x0008			#Indicates a security descriptor with a default DACL. For example, if the creator an object does not specify a DACL, the object receives the default DACL from the access token of the creator. This flag can affect how the system treats the DACL with respect to ACE inheritance. The system ignores this flag if the SE_DACL_PRESENT flag is not set.
										#This flag is used to determine how the final DACL on the object is to be computed and is not stored physically in the security descriptor control of the securable object.
										#To set this flag, use the SetSecurityDescriptorDacl function.
	SE_DACL_PRESENT = 0x0004			#Indicates a security descriptor that has a DACL. If this flag is not set, or if this flag is set and the DACL is NULL, the security descriptor allows full access to everyone.
										#This flag is used to hold the security information specified by a caller until the security descriptor is associated with a securable object. After the security descriptor is associated with a securable object, the SE_DACL_PRESENT flag is always set in the security descriptor control.
										#To set this flag, use the SetSecurityDescriptorDacl function.
	SE_DACL_PROTECTED = 0x1000			#Prevents the DACL of the security descriptor from being modified by inheritable ACEs. To set this flag, use the SetSecurityDescriptorControl function.
	SE_GROUP_DEFAULTED = 0x0002			#Indicates that the security identifier (SID) of the security descriptor group was provided by a default mechanism. This flag can be used by a resource manager to identify objects whose security descriptor group was set by a default mechanism. To set this flag, use the SetSecurityDescriptorGroup function.
	SE_OWNER_DEFAULTED = 0x0001			#Indicates that the SID of the owner of the security descriptor was provided by a default mechanism. This flag can be used by a resource manager to identify objects whose owner was set by a default mechanism. To set this flag, use the SetSecurityDescriptorOwner function.
	SE_RM_CONTROL_VALID = 0x4000		#Indicates that the resource manager control is valid.
	SE_SACL_AUTO_INHERIT_REQ = 0x0200	#Indicates a required security descriptor in which the system access control list (SACL) is set up to support automatic propagation of inheritable ACEs to existing child objects.
										#The system sets this bit when it performs the automatic inheritance algorithm for the object and its existing child objects. To convert a security descriptor and set this flag, protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function.
	SE_SACL_AUTO_INHERITED = 0x0800		#Indicates a security descriptor in which the system access control list (SACL) is set up to support automatic propagation of inheritable ACEs to existing child objects.
										#The system sets this bit when it performs the automatic inheritance algorithm for the object and its existing child objects. To convert a security descriptor and set this flag, protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function.
	SE_SACL_DEFAULTED = 0x0008			#A default mechanism, rather than the original provider of the security descriptor, provided the SACL. This flag can affect how the system treats the SACL, with respect to ACE inheritance. The system ignores this flag if the SE_SACL_PRESENT flag is not set. To set this flag, use the SetSecurityDescriptorSacl function.
	SE_SACL_PRESENT = 0x0010			#Indicates a security descriptor that has a SACL. To set this flag, use the SetSecurityDescriptorSacl function.
	SE_SACL_PROTECTED = 0x2000			#Prevents the SACL of the security descriptor from being modified by inheritable ACEs. To set this flag, use the SetSecurityDescriptorControl function.
	SE_SELF_RELATIVE = 0x8000			#Indicates a self-relative security descriptor. If this flag is not set, the security descriptor is in absolute format. For more information, see Absolute and Self-Relative Security Descriptors.



#https://docs.microsoft.com/en-us/windows/desktop/api/winnt/ns-winnt-_security_descriptor
class SECURITY_DESCRIPTOR:
	def __init__(self):
		self.Revision = None
		self.Sbz1 = None
		self.Control = None
		self.Owner = None
		self.Group = None
		self.Sacl = None
		self.Dacl = None
	
	@staticmethod
	def from_bytes(data):
		return SECURITY_DESCRIPTOR.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		sd = SECURITY_DESCRIPTOR()
		sd.Revision = int.from_bytes(buff.read(1), 'little', signed = False)
		sd.Sbz1 =  int.from_bytes(buff.read(1), 'little', signed = False)
		sd.Control = SE_SACL(int.from_bytes(buff.read(2), 'little', signed = False))
		OffsetOwner  = int.from_bytes(buff.read(4), 'little', signed = False)
		OffsetGroup  = int.from_bytes(buff.read(4), 'little', signed = False)
		OffsetSacl  = int.from_bytes(buff.read(4), 'little', signed = False)
		OffsetDacl  = int.from_bytes(buff.read(4), 'little', signed = False)
		if OffsetOwner > 0:
			buff.seek(OffsetOwner)
			sd.Owner = SID.from_buffer(buff)
		
		if OffsetGroup > 0:
			buff.seek(OffsetGroup)
			sd.Group = SID.from_buffer(buff)
			
		if OffsetSacl > 0:
			buff.seek(OffsetSacl)
			sd.Sacl = ACL.from_buffer(buff)
		
		if OffsetDacl > 0:
			buff.seek(OffsetDacl)
			sd.Dacl = ACL.from_buffer(buff)
			
		return sd
			
			
	def __str__(self):
		t = '=== SECURITY_DESCRIPTOR ==\r\n'
		t+= 'Revision : %s\r\n' % self.Revision
		t+= 'Control : %s\r\n' % self.Control
		t+= 'Owner : %s\r\n' % self.Owner
		t+= 'Group : %s\r\n' % self.Group
		t+= 'Sacl : %s\r\n' % self.Sacl
		t+= 'Dacl : %s\r\n' % self.Dacl
		return t



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# This is a failed attempt to simplify the SSPI integration.
# Sadly this doesnt work, as windows doesnt give the session key for some reason?
#

from winsspi.sspi import NegotiateSSPI, SSPIResult

class SPNEGO_SSPI:
	def __init__(self, settings):
		self.mode = 'CLIENT'
		self.settings = settings
		self.sspi = None
		self.username = None
		self.password = None
		self.target = None
		
		self.setup()
		
	def setup(self):
		if 'mode' in self.settings:
			self.mode = self.settings['mode']
		
		if 'username' in self.settings:
			self.username = self.settings['username']
			if 'password' in self.settings:
				self.password = self.settings['password']
				
		if 'target' in self.settings:
			self.username = self.settings['target']
	
		self.sspi = NegotiateSSPI()
		self.sspi.authGSSClientInit(self.target, client_name = self.username)
	
	def get_session_key(self):
		return self.sspi.get_session_key()
		
	async def encrypt(self, data, message_no):
		return await self.sspi.encrypt(data, message_no)

	async def decrypt(self, data, message_no):
		return await self.sspi.decrypt(data, message_no)
		
	async def authenticate(self, token, flags = None, seq_number = 0, is_rpc = False):
		try:
			if self.mode.upper() == 'CLIENT':
				res, data = self.sspi.authGSSClientStep(token)
				if res == SSPIResult.OK:
					return data[0][1], True
				elif res == SSPIResult.CONTINUE:
					return data[0][1], False
				else:
					raise Exception('SSPI errors')
					
			else:
				raise Exception('SERVER is not supported now')
		except Exception as e:
			import traceback
			traceback.print_exc()
	


import enum
import asyncio
import hmac
import hashlib
import platform
import copy
import traceback
import datetime

from aiosmb import logger
from aiosmb.commons.exceptions import *
from aiosmb.network.selector import NetworkSelector
from aiosmb.transport.netbios import NetBIOSTransport
from aiosmb.protocol.smb.command_codes import SMBCommand
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.protocol.smb.header import SMBHeader, SMBHeaderFlags2Enum
from aiosmb.protocol.smb.message import SMBMessage
from aiosmb.protocol.smb.commands import *
from aiosmb.protocol.smb2.message import SMB2Message, SMB2Transform, SMB2Compression
from aiosmb.protocol.smb2.commands.negotiate import SMB2ContextType, SMB2PreauthIntegrityCapabilities, SMB2HashAlgorithm, SMB2Cipher, SMB2CompressionType, SMB2CompressionFlags, SMB2EncryptionCapabilities, SMB2CompressionCapabilities
from aiosmb.protocol.smb2.commands import *
from aiosmb.protocol.smb2.headers import *
from aiosmb.protocol.smb2.command_codes import *
from aiosmb.protocol.common import *
from aiosmb.wintypes.dtyp.constrcuted_security.guid import *
from aiosmb.wintypes.access_mask import *
from aiosmb.wintypes.fscc.structures.fileinfoclass import *
from aiosmb.wintypes.fscc.structures.FileFullDirectoryInformation import *
from aiosmb.wintypes.fscc.FileAttributes import FileAttributes

from aiosmb.wintypes.dtyp.constrcuted_security.security_descriptor import SECURITY_DESCRIPTOR

from aiosmb.commons.smbcontainer import *
from aiosmb.commons.connection.target import *

from aiosmb.crypto.pure.AES.AESCCM import aesCCMEncrypt, aesCCMDecrypt
#from aiosmb.crypto.AESCCM_dome import aesCCMEncrypt, aesCCMDecrypt
from aiosmb.crypto.BASE import cipherMODE
from aiosmb.crypto.from_impacket import KDF_CounterMode, AES_CMAC
from aiosmb.crypto.compression.lznt1 import compress as lznt1_compress
from aiosmb.crypto.compression.lznt1 import decompress as lznt1_decompress

class SMBConnectionStatus(enum.Enum):
	NEGOTIATING = 'NEGOTIATING'
	SESSIONSETUP = 'SESSIONSETUP'
	RUNNING = 'RUNNING'
	CLOSED = 'CLOSED'
	
class TreeEntry:
	def __init__(self):
		self.share_name = None
		self.tree_id = None
		self.session_id = None
		self.number_of_users = None
		self.is_DFS = None
		self.is_CA = None
		self.is_scaleout = None
		self.encrypt = None
		self.maximal_access = None
		
	@staticmethod
	def from_tree_reply(reply, share_name):
		te = TreeEntry()
		te.share_name = share_name
		te.tree_id = reply.header.TreeId
		te.session_id = reply.header.SessionId
		te.number_of_users = 1
		te.is_DFS = TreeCapabilities.SMB2_SHARE_CAP_DFS in reply.command.Capabilities
		te.is_CA = TreeCapabilities.SMB2_SHARE_CAP_CONTINUOUS_AVAILABILITY in reply.command.Capabilities
		te.is_scaleout = TreeCapabilities.SMB2_SHARE_CAP_SCALEOUT in reply.command.Capabilities
		te.encrypt = ShareFlags.SMB2_SHAREFLAG_ENCRYPT_DATA in reply.command.ShareFlags
		te.maximal_access = reply.command.MaximalAccess 
		return te
	
class FileHandle:
	def __init__(self):
		self.file_id = None
		self.tree_id = None
		self.oplock_level = None
		self.is_durable = None
		self.is_resilient = None
		self.last_disconnect_time = None
		self.file_name = None
	
	@staticmethod	
	def from_create_reply(reply, tree_id, file_name, oplock_level):
		fh = FileHandle()
		fh.file_id = reply.command.FileId
		fh.tree_id = tree_id
		fh.oplock_level = oplock_level
		fh.is_durable = False
		fh.is_resilient = False
		fh.last_disconnect_time = 0
		fh.file_name = file_name
		return fh

class SMBPendingMsg:
	def __init__(self, message_id, OutstandingResponses, OutstandingResponsesEvent, timeout = 5, max_renewal = None):
		self.message_id = message_id
		self.max_renewal = max_renewal
		self.timeout = timeout #different operations require different timeouts, even depending on dielacts!!!
		self.OutstandingResponses = OutstandingResponses
		self.OutstandingResponsesEvent = OutstandingResponsesEvent
		self.pending_task = None

	async def __pending_waiter(self):
		await asyncio.sleep(self.timeout)
		await self.__destroy_message(SMBPendingTimeout())

	async def __destroy_message(self, problem):
		self.OutstandingResponses[self.message_id] = problem
		if self.message_id in self.OutstandingResponsesEvent:
			self.OutstandingResponsesEvent[self.message_id].set()
		return

	async def update(self):
		if self.pending_task is not None:
			self.pending_task.cancel()
		
		if self.max_renewal is not None:
			self.max_renewal -= 1
			if self.max_renewal == 0:
				await self.__destroy_message(SMBPendingMaxRenewal())
		self.pending_task = asyncio.create_task(self.__pending_waiter())


	async def run(self):
		self.pending_task = asyncio.create_task(self.__pending_waiter())

	async def stop(self):
		if self.pending_task is not None:
			self.pending_task.cancel()

class SMBConnection:
	"""
	Connection class for network connectivity and SMB messages management (sending/recieveing/singing/encrypting).
	"""
	#def __init__(self, gssapi, target, dialects = [NegotiateDialects.SMB202]):
	def __init__(self, gssapi, target):
		self.gssapi = gssapi
		self.original_gssapi = copy.deepcopy(gssapi) #preserving a copy of the original
		
		self.target = target
		
		#######DONT CHANGE THIS
		#use this for smb2 > self.supported_dialects = [NegotiateDialects.WILDCARD, NegotiateDialects.SMB202, NegotiateDialects.SMB210]
		#self.supported_dialects = [NegotiateDialects.SMB202, NegotiateDialects.SMB210]
		self.supported_dialects = self.target.preferred_dialects #[NegotiateDialects.WILDCARD, NegotiateDialects.SMB311]
		#######
		
		self.settings = None
		self.network_transport = None 
		self.netbios_transport = None #this class is used by the netbios transport class, keeping it here also maybe you like to go in raw
		self.incoming_task = None
		self.keepalive_task = None
		# TODO: turn it back on 
		self.supress_keepalive = True
		self.activity_at = None
		
		self.selected_dialect = None
		self.signing_required = False
		self.encryption_required = False
		
		self.status = SMBConnectionStatus.NEGOTIATING
		
		self.OutstandingResponsesEvent = {}
		self.OutstandingRequests = {}
		self.OutstandingResponses = {}

		self.pending_table = {}
		
		#two dicts for the same data, but with different lookup key
		self.TreeConnectTable_id = {}
		self.TreeConnectTable_share = {}
		
		self.FileHandleTable = {}
		
		self.SequenceWindow = 0
		self.MaxTransactSize = 0
		self.MaxReadSize = 0
		self.MaxWriteSize = 0
		self.ServerGuid = None
		self.RequireSigning = False
		self.ServerName = None
		self.ClientGUID = GUID.random()
		
		#self.Dialect = 0
		self.SupportsFileLeasing = False
		self.SupportsMultiCredit = False
		
		self.SupportsDirectoryLeasing = False
		self.SupportsMultiChannel = False
		self.SupportsPersistentHandles = False
		self.SupportsEncryption = False
		self.ClientCapabilities = 0
		self.ServerCapabilities = 0
		self.ClientSecurityMode = NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED | NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_REQUIRED
		self.ServerSecurityMode = 0
		
		
		self.SessionId = 0
		self.SessionKey = None
		self.SigningKey      = None
		self.ApplicationKey  = None
		self.EncryptionKey   = None
		self.DecryptionKey   = None

		
		

		### SMB311 global
		self.CompressAllRequests = False

		### SMB311 transport
		self.PreauthIntegrityHashId = SMB2HashAlgorithm.SHA_512 #the type of the hash function was negotiated , now there is only one type
		self.PreauthIntegrityHashValue = b'\x00'*64 #The preauthentication integrity hash value that was computed for the exchange of SMB2 NEGOTIATE request and response messages on this connection.
		self.CompressionId = None #the selected compression
		self.CipherId = None #
		self.CompressionIds = [SMB2CompressionType.LZNT1]#[SMB2CompressionType.NONE] #list of supported compression
		self.SupportsChainedCompression = False
		self.smb2_supported_encryptions = [SMB2Cipher.AES_128_CCM, SMB2Cipher.AES_128_GCM]
		
		self.preauth_ctx = hashlib.sha512

		#ignore_close is there to skip the logoff/closing of the channel
		#this is useful because there could be certain errors after a scusessful logon
		#that invalidates the whole session (eg. STATUS_USER_SESSION_DELETED)
		#if this happens then logoff will fail as well!
		self.session_closed = False 
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await asyncio.wait_for(self.terminate(), timeout = 1)

	def get_extra_info(self):
		try:
			ntlm_data = self.gssapi.get_extra_info().to_dict()
		except:
			traceback.print_exc()
			ntlm_data = None
		if self.ServerSecurityMode is not None:
			return {
				'ntlm_data' : ntlm_data,
				'signing_enabled' : NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED in self.ServerSecurityMode,
				'signing_required' : NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_REQUIRED in self.ServerSecurityMode,
			}
		return None

	async def __pending_task(self, message_id, timeout = 5):
		await asyncio.sleep(timeout)
		

	async def __handle_smb_in(self):
		"""
		Waits from SMB message bytes from the transport in_queue, and fills the connection table.
		This function started automatically when calling connect.
		Pls don't touch it.
		"""
		try:
			while True:
				msg_data, err = await self.netbios_transport.in_queue.get()
				self.activity_at = datetime.datetime.utcnow()
				if err is not None:
					logger.error('__handle_smb_in got error from transport layer %s' % err)
					#setting all outstanding events to finished
					for mid in self.OutstandingResponsesEvent:
						self.OutstandingResponses[mid] = None, err
						self.OutstandingResponsesEvent[mid].set()

					await self.terminate()
					return

				if msg_data[0] < 252:
					raise Exception('Unknown SMB packet type %s' % msg_data[0])

				if msg_data[0] == 0xFD:
					#encrypted transform
					msg = SMB2Transform.from_bytes(msg_data)
					
					if msg.header.EncryptionAlgorithm == SMB2Cipher.AES_128_CCM:
						dec_data = aesCCMDecrypt(msg.data, msg_data[20:], self.DecryptionKey, msg.header.Nonce[:11], msg.header.Signature)
						#print('dec_data %s' % dec_data)
						#cipher = AES(self.DecryptionKey, mode = cipherMODE.CCM, nonce = msg.header.Nonce[:11])
						#cipher.update(msg_data[20:])
						#dec_data = cipher.decrypt(msg.data)
						#calc_signature = cipher.verify()


						#cipher = AES.new(self.DecryptionKey, AES.MODE_CCM, msg.header.Nonce[:11])
						#cipher.update(msg_data[20:])
						#dec_data = cipher.decrypt(msg.data)
						##calc_signature = cipher.verify()

						# TODO: add signature checking!!!!!

						msg_data = dec_data
					
					else:
						raise Exception('Common encryption algo is %s but it is not implemented!' % msg.header.EncryptionAlgorithm)

				if msg_data[0] == 0xFC:
					#compressed transform
					msg = SMB2Compression.from_bytes(msg_data)
					if msg.header.Flags == SMB2CompressionFlags.NONE:
						if msg.header.CompressionAlgorithm != self.CompressionId:
							logger.debug('Server is using a different compression algo than whats agreed upon...')
						if msg.header.CompressionAlgorithm == SMB2CompressionType.LZNT1:
							uncompressed_data = msg.data[:msg.header.Offset]
							uncompressed_data += lznt1_decompress(msg.data[msg.header.Offset:])
							msg_data = uncompressed_data
						else:
							raise Exception('Server used %s compression, but it is not implemented' % msg.header.CompressionAlgorithm.name)
					else:
						raise Exception('Server sent chained compression, but its not implemented here')
					
				if msg_data[0] == 0xFE:
					#version2
					msg = SMB2Message.from_bytes(msg_data)

				if msg_data[0] == 0xFF:
					#version1
					msg = SMBMessage.from_bytes(msg_data)
					

				logger.log(1, '__handle_smb_in got new message with Id %s' % msg.header.MessageId)
				
				if isinstance(msg, SMB2Transform):
					#message is encrypted
					#this point we should decrypt it and only store the decrypted part in the OutstandingResponses table
					#but for now we just thropw exception bc encryption is not implemented
					raise Exception('Encrypted SMBv2 message recieved, but encryption is not yet supported!')
				
				if msg.header.Status == NTStatus.PENDING:
					self.pending_table[msg.header.MessageId] = SMBPendingMsg(msg.header.MessageId, self.OutstandingResponses, self.OutstandingResponsesEvent)
					await self.pending_table[msg.header.MessageId].run()
					continue

				if msg.header.MessageId in self.pending_table:
					await self.pending_table[msg.header.MessageId].stop()
					del self.pending_table[msg.header.MessageId]

				self.OutstandingResponses[msg.header.MessageId] = (msg, msg_data)
				if msg.header.MessageId in self.OutstandingResponsesEvent:
					self.OutstandingResponsesEvent[msg.header.MessageId].set()
				else:

					#here we are loosing messages, the functionality for "PENDING" and "SHARING_VIOLATION" should be implemented
					continue
		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
		except:
			logger.exception('__handle_smb_in')
			
			
	async def login(self):
		"""
		This is the normal starting function.
		Performs establishment of the TCP connection, then the negotiation and finally the session setup.
		If this function returns without an exception, then I'm happy.
		Also it means that you have a working and active session to the server.
		"""
		results = await asyncio.gather(*[self.connect()], return_exceptions=True)
		if isinstance(results[0], Exception):
			raise results[0]
		await self.negotiate()
		await self.session_setup()
		
		#TODO: REENABLE THIS!
		self.keepalive_task = asyncio.create_task(self.keepalive())
		
	async def fake_login(self):
		if 'NTLMSSP - Microsoft NTLM Security Support Provider' not in self.gssapi.authentication_contexts:
			raise Exception('Fake authentication is only supported via NTLM package')
		await self.connect()
		await self.negotiate()
		await self.session_setup(fake_auth = True)
		await self.disconnect()
		return self.gssapi.get_extra_info()
		
	async def connect(self):
		"""
		Establishes socket connection to the remote endpoint. Also starts the internal reading procedures.
		"""
		self.network_transport = await NetworkSelector.select(self.target)
		
		res = await asyncio.gather(*[self.network_transport.connect()], return_exceptions=True)
		if isinstance(res[0], Exception):
			raise res[0]
		
		
		self.netbios_transport = NetBIOSTransport(self.network_transport)
		res =  await asyncio.gather(*[self.netbios_transport.run()], return_exceptions=True)
		if isinstance(res[0], Exception):
			raise res[0]
		
		self.incoming_task = asyncio.create_task(self.__handle_smb_in())
		
	async def disconnect(self):
		"""
		Tears down the socket connecting as well as the reading cycle.
		Doesn't do any cleanup! 
		For proper cleanup call the terminate function.
		"""
		if self.status == SMBConnectionStatus.CLOSED:
			return
		
		self.status = SMBConnectionStatus.CLOSED
		if self.netbios_transport:
			await self.netbios_transport.stop()
		if self.network_transport:
			await self.network_transport.disconnect()
		if self.incoming_task:
			self.incoming_task.cancel()
		
		if self.keepalive_task:
			self.keepalive_task.cancel()

	async def keepalive(self):
		"""
		Sends an echo message every X seconds to the server to keep the channel open
		"""
		try:
			sleep_time = 10
			if self.target.timeout < 0:
				return
			elif self.target.timeout > 0:
				sleep_time = max(self.target.timeout - 1, sleep_time)
			while True:
				await asyncio.sleep(sleep_time)
				if (datetime.datetime.utcnow() - self.activity_at).seconds > sleep_time and self.supress_keepalive is False: 
					await self.echo()
				
		except asyncio.CancelledError:
			return
		except Exception as e:
			logger.error('Keepalive failed! Server probably disconnected!')
			await self.disconnect()

	def update_integrity(self, msg_data):
		#if is_neg is True:
		#	self.PreauthIntegrityHashValue = b'\x00'*64
		#print('update_integrity with data : %s' % msg_data)
		ctx = hashlib.sha512()
		ctx.update(self.PreauthIntegrityHashValue + msg_data)
		self.PreauthIntegrityHashValue = ctx.digest()
		
	async def negotiate(self):
		"""
		Initiates protocol negotiation.
		First we send an SMB_COM_NEGOTIATE_REQ with our supported dialects
		"""
		rply = None
		if NegotiateDialects.WILDCARD in self.supported_dialects:
			###let's construct an SMBv1 SMB_COM_NEGOTIATE_REQ packet
			header = SMBHeader()
			header.Command  = SMBCommand.SMB_COM_NEGOTIATE
			header.Status   = NTStatus.SUCCESS
			header.Flags    = 0
			header.Flags2   = SMBHeaderFlags2Enum.SMB_FLAGS2_UNICODE
				
			command = SMB_COM_NEGOTIATE_REQ()				
			command.Dialects = ['SMB 2.???','SMB 2.002']
			
			msg = SMBMessage(header, command)
			message_id = await self.sendSMB(msg)
			#recieveing reply, should be version2, because currently we dont support v1 :(
			rply, rply_data = await self.recvSMB(message_id, ret_data = True) #negotiate MessageId should be 1
			if isinstance(rply, SMBMessage):
				raise Exception('Server replied with SMBv1 message, doesnt support SMBv2')
			if rply.header.Status != NTStatus.SUCCESS:
				raise Exception('SMB2 negotiate error! Server replied with error status code!')
			
			del self.supported_dialects[NegotiateDialects.WILDCARD]

			#print(rply.command.Capabilities)

		command = NEGOTIATE_REQ()
		command.SecurityMode    = self.ClientSecurityMode
		command.Capabilities    = 0
		command.ClientGuid      = self.ClientGUID
		command.Dialects        = [dialect for dialect in self.supported_dialects]

		#if all(dialect in SMB2_NEGOTIATE_DIALTECTS_2 for dialect in self.supported_dialects):
		#	command.Capabilities    = NegotiateCapabilities.ENCRYPTION
		if all(dialect in SMB2_NEGOTIATE_DIALTECTS_3 for dialect in self.supported_dialects):
			command.Capabilities    = NegotiateCapabilities.ENCRYPTION


		if NegotiateDialects.SMB311 in self.supported_dialects:
			#SMB311 mandates the contextlist to be populated
						
			command.NegotiateContextList.append(
				SMB2PreauthIntegrityCapabilities.construct(
					[self.PreauthIntegrityHashId]
				)
			)
						
			if self.smb2_supported_encryptions is not None:
				command.Capabilities |= NegotiateCapabilities.ENCRYPTION
				command.NegotiateContextList.append(
					SMB2EncryptionCapabilities.from_enc_list(
						self.smb2_supported_encryptions
					)
				)

			if self.CompressionIds is not None:
				command.NegotiateContextList.append(
					SMB2CompressionCapabilities.from_comp_list(
						self.CompressionIds,
						self.SupportsChainedCompression
					)
				)
		#print('aaaaa')
		#input(command.NegotiateContextList)
		#print('aaaaa')
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.NEGOTIATE
		header.CreditReq = 0
					
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg)
		rply, rply_data = await self.recvSMB(message_id, ret_data=True) #negotiate MessageId should be 1
		if isinstance(rply, SMBMessage):
			raise Exception('Server replied with SMBv1 message, doesnt support SMBv2')
		#print('NEG RPLY updates')
		self.update_integrity(rply_data)
		
		if rply.header.Status != NTStatus.SUCCESS:
			if rply.header.Status == NTStatus.NOT_SUPPORTED:
				raise Exception('negotiate_1 dialect probably not suppported by the server. reply: %s' % rply.header.Status)
			else:
				raise Exception('negotiate_1 reply: %s' % rply.header.Status)
					
		if rply.command.DialectRevision not in self.supported_dialects:
			raise SMBUnsupportedDialectSelected()
				
		self.selected_dialect = rply.command.DialectRevision
		self.ServerSecurityMode = rply.command.SecurityMode
		self.signing_required = NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED in rply.command.SecurityMode
				
		if NegotiateCapabilities.ENCRYPTION in rply.command.Capabilities:
			self.encryption_required = True
			self.CipherId = SMB2Cipher.AES_128_CCM

		for negctx in rply.command.NegotiateContextList:
			if negctx.ContextType == SMB2ContextType.ENCRYPTION_CAPABILITIES:
				self.encryption_required = True
				self.CipherId = negctx.Ciphers[0]
					
			if negctx.ContextType == SMB2ContextType.COMPRESSION_CAPABILITIES:
				self.CompressionId = negctx.CompressionAlgorithms[0]

		logger.log(1, 'Server selected dialect: %s' % self.selected_dialect)
				
		self.MaxTransactSize = min(0x100000, rply.command.MaxTransactSize)
		self.MaxReadSize = min(0x100000, rply.command.MaxReadSize)
		self.MaxWriteSize = min(0x100000, rply.command.MaxWriteSize)
		self.ServerGuid = rply.command.ServerGuid
		self.SupportsMultiChannel = NegotiateCapabilities.MULTI_CHANNEL in rply.command.Capabilities
		self.SupportsFileLeasing = NegotiateCapabilities.LEASING in rply.command.Capabilities
		self.SupportsMultiCredit = NegotiateCapabilities.LARGE_MTU in rply.command.Capabilities
				
		self.ClientCapabilities = rply.command.Capabilities
		self.ServerCapabilities = rply.command.Capabilities
		#self.ClientSecurityMode = 0			
		self.status = SMBConnectionStatus.SESSIONSETUP

	async def session_setup(self, fake_auth = False):
		
		authdata = None
		status = NTStatus.MORE_PROCESSING_REQUIRED
		maxiter = 5
		while status == NTStatus.MORE_PROCESSING_REQUIRED and maxiter > 0:
			command = SESSION_SETUP_REQ()
			try:
				command.Buffer, res  = await self.gssapi.authenticate(authdata)
				if fake_auth == True:
					if self.gssapi.selected_authentication_context is not None and self.gssapi.selected_authentication_context.ntlmChallenge is not None:
						return
			except Exception as e:
				logger.exception('GSSAPI auth failed!')
				#TODO: Clear this up, kerberos lib needs it's own exceptions!
				if str(e).find('Preauth') != -1:
					raise SMBKerberosPreauthFailed()
				else:
					raise e
					#raise SMBKerberosPreauthFailed()
			
			command.Flags = 0
			command.SecurityMode = NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED
			command.Capabilities = 0 #self.ClientCapabilities
			command.Channel      = 0
			command.PreviousSessionId    = 0
			
			header = SMB2Header_SYNC()
			header.Command  = SMB2Command.SESSION_SETUP
			header.CreditReq = 0
			
			msg = SMBMessage(header, command)
			message_id = await self.sendSMB(msg)
			#self.update_integrity(sent_msg.to_bytes())
			rply, rply_data = await self.recvSMB(message_id, ret_data=True)
			
			if self.SessionId == 0:
				self.SessionId = rply.header.SessionId
			
			if rply.header.Status not in [NTStatus.SUCCESS, NTStatus.MORE_PROCESSING_REQUIRED]:
				break

			if rply.header.Status != NTStatus.SUCCESS:
				print('RPLY updates')
				self.update_integrity(rply_data)
			
			authdata = rply.command.Buffer
			status = rply.header.Status
			maxiter -= 1
		
		if rply.header.Status == NTStatus.SUCCESS:
			self.SessionKey = self.gssapi.get_session_key()[:16]
			
			# TODO: key calc
			if self.signing_required and self.selected_dialect in [NegotiateDialects.SMB300 , NegotiateDialects.SMB302 , NegotiateDialects.SMB311]:
				if  self.selected_dialect == NegotiateDialects.SMB311:
					#SMB311 is a special snowflake
					self.SigningKey      = KDF_CounterMode(self.SessionKey, b"SMBSigningKey\x00", self.PreauthIntegrityHashValue, 128)
					self.ApplicationKey  = KDF_CounterMode(self.SessionKey, b"SMBAppKey\x00", self.PreauthIntegrityHashValue, 128)
					self.EncryptionKey   = KDF_CounterMode(self.SessionKey, b"SMBC2SCipherKey\x00", self.PreauthIntegrityHashValue, 128)
					self.DecryptionKey   = KDF_CounterMode(self.SessionKey, b"SMBS2CCipherKey\x00", self.PreauthIntegrityHashValue, 128)
				else:
					self.SigningKey      = KDF_CounterMode(self.SessionKey, b"SMB2AESCMAC\x00", b"SmbSign\x00", 128)
					self.ApplicationKey  = KDF_CounterMode(self.SessionKey, b"SMB2APP\x00", b"SmbRpc\x00", 128)
					self.EncryptionKey   = KDF_CounterMode(self.SessionKey, b"SMB2AESCCM\x00", b"ServerIn \x00", 128)
					self.DecryptionKey   = KDF_CounterMode(self.SessionKey, b"SMB2AESCCM\x00", b"ServerOut\x00", 128)
			
			self.status = SMBConnectionStatus.RUNNING
		
		elif rply.header.Status == NTStatus.LOGON_FAILURE:
			raise SMBAuthenticationFailed()
		
		else:
			raise SMBException('session_setup (authentication probably failed)', rply.header.Status)
			
		
		
	async def recvSMB(self, message_id, ret_data = False):
		"""
		Returns an SMB message from the outstandingresponse dict, OR waits until the expected message_id appears.
		"""
		if message_id not in self.OutstandingResponses:
			logger.log(1, 'Waiting on messageID : %s' % message_id)
			await self.OutstandingResponsesEvent[message_id].wait()
			
		msg, msg_data = self.OutstandingResponses.pop(message_id)
		if msg is None:
			# this indicates and exception, so the msg_data is the exception
			raise msg_data
		
		if self.status != SMBConnectionStatus.NEGOTIATING:
			if self.selected_dialect != NegotiateDialects.SMB202:
				self.SequenceWindow += (msg.header.CreditCharge - 1)

		if msg.header.Status != NTStatus.PENDING:
			if message_id in self.OutstandingResponsesEvent:
				del self.OutstandingResponsesEvent[message_id]
		else:
			self.OutstandingResponsesEvent[message_id].clear()
		
		if ret_data is False:
			return msg
		return msg, msg_data
		
	def sign_message(self, msg):
		if self.selected_dialect in [NegotiateDialects.SMB202, NegotiateDialects.SMB210]:
			if self.SessionKey:
				msg.header.Flags = msg.header.Flags ^ SMB2HeaderFlag.SMB2_FLAGS_SIGNED ##maybe move this flag settings to sendsmb since singing is determined there?
				digest = hmac.new(self.SessionKey, msg.to_bytes(), hashlib.sha256).digest()
				msg.header.Signature = digest[:16]
		else:
			if self.SigningKey:
				msg_data = msg.to_bytes()	
				signature = AES_CMAC(self.SigningKey, msg_data, len(msg_data))
				msg.header.Signature = signature

	
	def encrypt_message(self, msg_data):
		nonce = os.urandom(11)

		hdr = SMB2Header_TRANSFORM()
		#hdr.Signature = None
		hdr.Nonce = nonce + (b'\x00' * 5) 
		hdr.OriginalMessageSize = len(msg_data)
		hdr.EncryptionAlgorithm = SMB2Cipher.AES_128_CCM
		hdr.SessionId = self.SessionId

		enc_data, hdr.Signature = aesCCMEncrypt(msg_data, hdr.to_bytes()[20:], self.EncryptionKey, nonce)

		#cipher = AES.new(self.EncryptionKey, AES.MODE_CCM, nonce)
		#cipher.update(hdr.to_bytes()[20:])
		#enc_data = cipher.encrypt(msg_data)
		#hdr.Signature = cipher.digest()
		return SMB2Transform(hdr, enc_data)


	def compress_message(self, msg):
		print('COMPRESSION')
		msg_data = msg.to_bytes()
		
		if self.SupportsChainedCompression is False:
			if self.CompressionId == SMB2CompressionType.NONE:
				compressed_data = msg_data
			elif self.CompressionId == SMB2CompressionType.LZNT1:
				compressed_data = lznt1_compress(msg_data)
			else:
				raise Exception('Common compression type is %s but its not implemented!' % self.CompressionId)
			
			comp_hdr = SMB2Header_COMPRESSION_TRANSFORM()
			comp_hdr.OriginalCompressedSegmentSize = len(msg_data)
			comp_hdr.CompressionAlgorithm = self.CompressionId
			comp_hdr.Flags = SMB2CompressionFlags.NONE
			comp_hdr.Offset = 0 #it'z zero because we compress the full message

			return SMB2Compression(comp_hdr, compressed_data)
		else:
			raise Exception('Chained compression not implemented!')
		
	async def sendSMB(self, msg, ret_message = False, compression_cb = None):
		"""
		Sends an SMB message to teh remote endpoint.
		msg: SMB2Message or SMBMessage
		Returns: MessageId integer
		"""
		self.activity_at = datetime.datetime.utcnow()
		if self.status == SMBConnectionStatus.NEGOTIATING:
			if isinstance(msg, SMBMessage):
				#creating an event for outstanding response
				#self.OutstandingResponsesEvent[0] = asyncio.Event()
				#await self.netbios_transport.out_queue.put(msg.to_bytes())
				#self.SequenceWindow += 1
				#return 0
				message_id = 0
				self.SequenceWindow += 1
			else:
				msg.header.CreditCharge = 1
				msg.header.CreditReq = 0
				msg.header.MessageId = self.SequenceWindow
				message_id = self.SequenceWindow
				self.SequenceWindow += 1
				self.update_integrity(msg.to_bytes())

			self.OutstandingResponsesEvent[message_id] = asyncio.Event()
			await self.netbios_transport.out_queue.put(msg.to_bytes())
			if ret_message is True:
				return message_id, msg
			return message_id
				

		if msg.header.Command is not SMB2Command.CANCEL:
			msg.header.MessageId = self.SequenceWindow
			self.SequenceWindow += 1
		
		msg.header.SessionId = self.SessionId
		
		if not msg.header.CreditCharge:
			msg.header.CreditCharge = 1

		
		
		if self.status != SMBConnectionStatus.SESSIONSETUP:
			msg.header.CreditReq = 127
		
		message_id = msg.header.MessageId
		

		if self.CompressionId is not None and self.EncryptionKey is not None:
			if compression_cb is None:
				msg = self.compress_message(msg)
			else:
				msg = compression_cb(msg)
		if self.signing_required == True:
			self.sign_message(msg)
		
		if self.encryption_required == True and self.EncryptionKey is not None:
			msg = self.encrypt_message(msg.to_bytes())

		else:
			#print('SEND updates')
			self.update_integrity(msg.to_bytes())

		#creating an event for outstanding response
		self.OutstandingResponsesEvent[message_id] = asyncio.Event()
		await self.netbios_transport.out_queue.put(msg.to_bytes())
		
		if ret_message is True:
				return message_id, msg
		return message_id
		
	async def tree_connect(self, share_name):
		"""
		share_name MUST be in "\\\\server\\share" format! Server can be NetBIOS name OR IP4 OR IP6 OR FQDN
		"""
		if self.session_closed == True:
			return
		command = TREE_CONNECT_REQ()
		command.Path = share_name
		command.Flags = 0
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.TREE_CONNECT
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)
		
		rply = await self.recvSMB(message_id)
		
		if rply.header.Status == NTStatus.SUCCESS:
			te = TreeEntry.from_tree_reply(rply, share_name)
			self.TreeConnectTable_id[rply.header.TreeId] = te
			self.TreeConnectTable_share[share_name] = te
			return te
		
		elif rply.header.Status == NTStatus.BAD_NETWORK_NAME:
			raise SMBIncorrectShareName()
			
		elif rply.header.Status == NTStatus.USER_SESSION_DELETED:
			self.session_closed = True
			raise SMBException('session delted', NTStatus.USER_SESSION_DELETED)
		
		
		else:
			raise SMBException('', rply.header.Status)
		
	async def create(self, tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs, impresonation_level = ImpersonationLevel.Impersonation, oplock_level = OplockLevel.SMB2_OPLOCK_LEVEL_NONE, create_contexts = None, return_reply = False):
		if self.session_closed == True:
			return
		
		if tree_id not in self.TreeConnectTable_id:
			raise Exception('Unknown Tree ID!')
		
		command = CREATE_REQ()
		command.RequestedOplockLevel  = oplock_level
		command.ImpersonationLevel  = impresonation_level
		command.DesiredAccess    = desired_access
		command.FileAttributes     = file_attrs
		command.ShareAccess      = share_mode
		command.CreateDisposition       = create_disposition
		command.CreateOptions        = create_options
		command.Name = file_path
		command.CreateContext = create_contexts
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.CREATE
		header.TreeId = tree_id
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)
		
		rply = await self.recvSMB(message_id)
		
		if rply.header.Status == NTStatus.SUCCESS:
			fh = FileHandle.from_create_reply(rply, tree_id, file_path, oplock_level)
			self.FileHandleTable[fh.file_id] = fh
			
			if return_reply == True:
				return rply.command.FileId, rply.command
			return rply.command.FileId
		
		elif rply.header.Status == NTStatus.ACCESS_DENIED:
			#this could mean incorrect filename/foldername OR actually access denied
			raise SMBCreateAccessDenied()
			
		else:
			raise SMBException('', rply.header.Status)
	
	async def read(self, tree_id, file_id, offset = 0, length = 0):
		"""
		Will issue one read command only then waits for reply. To read a whole file you must use a filereader logic! 
		returns the data bytes and the remaining data length
		
		IMPORTANT: remaning data length is dependent on the length of the requested chunk (length param) not on the actual file length.
		to get the remaining length for the actual file you must set the length parameter to the correct file size!
		
		If and EOF happens the function returns an empty byte array and the remaining data is set to 0
		"""
		if self.session_closed == True:
			return
			
		if tree_id not in self.TreeConnectTable_id:
			raise Exception('Unknown Tree ID!')
		if file_id not in self.FileHandleTable:
			raise Exception('Unknown File ID!')
			
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.READ
		header.TreeId = tree_id
		
		if length < self.MaxReadSize:
			length = self.MaxReadSize
		
		if self.selected_dialect != NegotiateDialects.SMB202 and self.SupportsMultiCredit == True:
			header.CreditCharge = ( 1 + (length - 1) // 65536)
		else: 
			length = min(65536,length)
			
		command = READ_REQ()
		command.Length = length
		command.Offset = offset
		command.FileId = file_id
		command.MinimumCount = 0
		command.RemainingBytes = 0
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)
		
		rply = await self.recvSMB(message_id)
		
		if rply.header.Status == NTStatus.SUCCESS:
			return rply.command.Buffer, rply.command.DataRemaining
		
		elif rply.header.Status == NTStatus.END_OF_FILE:
			return b'', 0
			
		else:
			raise SMBException('', rply.header.Status)
			
			
	async def write(self, tree_id, file_id, data, offset = 0):
		"""
		This function will send one packet only! The data size can be larger than what one packet allows, but it will be truncated
		to the maximum. 
		Also, there is no guarantee that the actual sent data will be fully written to the remote file! This will be indicated in the returned value.
		Use a high-level function to get a full write.
		
		"""
		if self.session_closed == True:
			return
			
		if tree_id not in self.TreeConnectTable_id:
			raise Exception('Unknown Tree ID!')
		if file_id not in self.FileHandleTable:
			raise Exception('Unknown File ID!')
			
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.WRITE
		header.TreeId = tree_id
			
		if len(data) > self.MaxWriteSize:
			data = data[:self.MaxWriteSize]
			
		if self.selected_dialect != NegotiateDialects.SMB202 and self.SupportsMultiCredit == True:
			header.CreditCharge = ( 1 + (len(data) - 1) // 65536)
		else: 
			data = data[:min(65536,len(data))]
		
		command = WRITE_REQ()
		command.Length = len(data)
		command.Offset = offset
		command.FileId = file_id
		command.Data = data
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)
		
		rply = await self.recvSMB(message_id)
		
		if rply.header.Status == NTStatus.SUCCESS:
			return rply.command.Count
		
		else:
			SMBException('', rply.header.Status)
		
	async def query_info(self, tree_id, file_id, info_type = QueryInfoType.FILE, information_class = FileInfoClass.FileStandardInformation, additional_information = 0, flags = 0, data_in = ''):
		"""
		Queires the file or directory for specific information. The information returned is depending on the input parameters, check the documentation on msdn for a better understanding.
		The resturned data can by raw bytes or an actual object, depending on wther your info is implemented in the library.
		Sorry there are a TON of classes to deal with :(
		
		IMPORTANT: in case you are requesting big amounts of data, the result will arrive in chunks. You will need to invoke this function until None is returned to get the full data!!!
		"""
		if self.session_closed == True:
			return
		if tree_id not in self.TreeConnectTable_id:
			raise Exception('Unknown Tree ID!')
		if file_id not in self.FileHandleTable:
			raise Exception('Unknown File ID!')
			
		command = QUERY_INFO_REQ()
		command.InfoType = info_type
		command.FileInfoClass = information_class
		command.AdditionalInformation = additional_information
		command.Flags = flags
		command.FileId = file_id
		command.Data = data_in
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.QUERY_INFO
		header.TreeId = tree_id
		
		msg = SMBMessage(header, command)
		message_id = await self.sendSMB(msg)

		rply = await self.recvSMB(message_id)
		
		if rply.header.Status == NTStatus.SUCCESS:
			#https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/3b1b3598-a898-44ca-bfac-2dcae065247f
			if info_type == QueryInfoType.SECURITY:
				return SECURITY_DESCRIPTOR.from_bytes(rply.command.Data)
				
			elif info_type == QueryInfoType.FILE:
				if information_class == FileInfoClass.FileFullDirectoryInformation:
					return FileFullDirectoryInformationList.from_bytes(rply.command.Data)
					
				else:
					return rply.command.Data
					
			elif info_type == QueryInfoType.FILESYSTEM:
				#TODO: implement this
				return rply.command.Data
				
			elif info_type == QueryInfoType.QUOTA:
				#TODO: implement this
				return rply.command.Data
			
			else:
				#this should never happen
				return rply.command.Data
			
		else:
			raise SMBException('', rply.header.Status)
			
		
	async def query_directory(self, tree_id, file_id, search_pattern = '*', resume_index = 0, information_class = FileInfoClass.FileFullDirectoryInformation, maxBufferSize = None, flags = 0):
		"""
		
		IMPORTANT: in case you are requesting big amounts of data, the result will arrive in chunks. You will need to invoke this function until None is returned to get the full data!!!
		"""
		if self.session_closed == True:
			return
			
		if tree_id not in self.TreeConnectTable_id:
			raise Exception('Unknown Tree ID!')
		if file_id not in self.FileHandleTable:
			raise Exception('Unknown File ID!')
			
		
		command = QUERY_DIRECTORY_REQ()
		command.FileInformationClass  = information_class
		command.Flags = 0
		if resume_index != 0 :
			command.Flags |= QueryDirectoryFlag.SMB2_INDEX_SPECIFIED
		command.FileIndex  = resume_index
		command.FileId  = file_id
		command.FileName = search_pattern
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.QUERY_DIRECTORY
		header.TreeId = tree_id
		
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg)

		rply = await self.recvSMB(message_id)
		
		if rply.header.Status == NTStatus.SUCCESS:
			if information_class == FileInfoClass.FileFullDirectoryInformation:
				return FileFullDirectoryInformationList.from_bytes(rply.command.Data)
				
			else:
				return rply.command.Data
				
		elif rply.header.Status == NTStatus.NO_MORE_FILES:
			return None
		else:
			raise Exception('query_directory reply: %s' % rply.header.Status)
			
	async def close(self, tree_id, file_id, flags = CloseFlag.NONE):
		"""
		Closes the file/directory/pipe/whatever based on file_id. It will automatically remove all traces of the file handle.
		"""
		if self.session_closed == True:
			return
		command = CLOSE_REQ()
		command.Flags = flags
		command.FileId = file_id
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.CLOSE
		header.TreeId = tree_id
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg)
		

		rply = await self.recvSMB(message_id)
		if rply.header.Status == NTStatus.SUCCESS:
			del self.FileHandleTable[file_id]
			
			
	async def flush(self, tree_id, file_id):
		"""
		Flushes all cached data that may be on the server for the given file.
		"""
		if self.session_closed == True:
			return
			
		command = FLUSH_REQ()
		command.FileId = file_id
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.FLUSH
		header.TreeId = tree_id
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg)

		rply = await self.recvSMB(message_id)
		
	async def logoff(self):
		"""
		Logs off from the server, effectively terminates the session. 
		The underlying connection will still be active, so please either clean it up manually or dont touch this function
		For proper closing of the connection use the terminate function
		"""
		if self.session_closed == True:
			return
			
		if self.status == SMBConnectionStatus.CLOSED:
			return
			
		command = LOGOFF_REQ()
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.LOGOFF
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg)

		rply = await self.recvSMB(message_id)
	
	async def echo(self):
		"""
		Issues an ECHO request to the server. Server will reply with and ECHO response, if it's still alive
		"""
		if self.session_closed == True:
			return
		command = ECHO_REQ()
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.ECHO
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg)
		rply = await self.recvSMB(message_id)
		
	async def tree_disconnect(self, tree_id):
		"""
		Disconnects from tree, removes all file entries associated to the tree
		"""
		if self.session_closed == True:
			return
			
		command = TREE_DISCONNECT_REQ()
		
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.TREE_DISCONNECT
		header.TreeId = tree_id
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg)

		rply = await self.recvSMB(message_id)
		
		if rply.header.Status == NTStatus.SUCCESS:
			del_file_ids = []
			share_name = self.TreeConnectTable_id[tree_id].share_name
			for fe in self.FileHandleTable:
				if self.FileHandleTable[fe].tree_id == tree_id:
					del_file_ids.append(self.FileHandleTable[fe].file_id)
			
			for file_id in del_file_ids:
				del self.FileHandleTable[file_id]
			
			del self.TreeConnectTable_id[tree_id]
			del self.TreeConnectTable_share[share_name]

		
	async def cancel(self, message_id):
		"""
		Issues a CANCEL command for the given message_id
		"""
		if self.session_closed == True:
			return
			
		command = CANCEL_REQ()
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.CANCEL
		msg = SMB2Message(header, command)
		msg.header.MessageId = message_id
		message_id = await self.sendSMB(msg)
		rply = await self.recvSMB(message_id)
		
	async def terminate(self):
		"""
		Use this function to properly terminate the SBM connection.
		Terminates the connection. Closes all tree handles, logs off and disconnects the TCP connection.
		"""
		#return
		try:
			logger.debug('Terminate called!')			
			if self.session_closed == True or self.status == SMBConnectionStatus.CLOSED:
				logger.debug('Terminate 1!')
				return
			
			if self.status == SMBConnectionStatus.RUNNING:
				logger.debug('Terminate 2!')
				#only doing the proper disconnection if the connection was already running
				for tree_id in list(self.TreeConnectTable_id.keys()):
					try:
						await asyncio.wait_for(self.tree_diconnect(tree_id), timeout = self.target.timeout)
					except:
						pass
				logger.debug('Terminate 3!')
				#logging off
				try:
					await asyncio.wait_for(self.logoff(), timeout = self.target.timeout)
				except Exception as e:
					pass
			
			logger.debug('Terminate 4!')
			#terminating TCP connection
			await asyncio.wait_for(self.disconnect(), timeout = self.target.timeout)
			logger.debug('Terminate finished!')	
		except:
			logger.exception('')
	
	async def ghosting(self):
		#self.encryption_required = False
		self.supress_keepalive = True
		def compress_callback(msg):
			print('Callback is here!')
			msg_data = msg.to_bytes()
			compressed_data = lznt1_compress(msg_data)
			comp_hdr = SMB2Header_COMPRESSION_TRANSFORM()
			comp_hdr.OriginalCompressedSegmentSize = len(msg_data)
			comp_hdr.CompressionAlgorithm = SMB2CompressionType.LZNT1 #self.CompressionId
			comp_hdr.Flags = SMB2CompressionFlags.NONE
			comp_hdr.Offset = 0 #this should be marking the start of the compressed data

			return SMB2Compression(comp_hdr, compressed_data)

		print('################################### TESTING compression ###################################')
		await self.echo()
		#command = ECHO_REQ()
		#header = SMB2Header_SYNC()
		#header.Command  = SMB2Command.ECHO
		#msg = SMB2Message(header, command)
		#message_id = await self.sendSMB(msg)
		#print('Waiting for reply')
		#rply = await self.recvSMB(message_id)
		#print(rply)
		print('################################### SPLOIT SPLOIT SPLOIT ################################')

		
		command = ECHO_REQ()
		header = SMB2Header_SYNC()
		header.Command  = SMB2Command.ECHO
		msg = SMB2Message(header, command)
		message_id = await self.sendSMB(msg, compression_cb = compress_callback)
		try:
			rply = await asyncio.wait_for(self.recvSMB(message_id), timeout = 0.5)
			return True
		except:
			return False


#async def test(target):
#	#setting up NTLM auth
#	template_name = 'Windows10_15063_knowkey'
#	credential = Credential()
#	credential.username = 'victim'
#	credential.password = 'Passw0rd!1'
#	credential.domain = 'TEST'
#	
#	settings = NTLMHandlerSettings(credential, mode = 'CLIENT', template_name = template_name)
#	handler = NTLMAUTHHandler(settings)
#	
#	#setting up SPNEGO
#	spneg = SPNEGO()
#	spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
#	connection = SMBConnection(spneg, [NegotiateDialects.SMB210])
#	await connection.connect(target)
#	await connection.negotiate()
#	await connection.session_setup()
#	tree_entry = await connection.tree_connect('\\\\10.10.10.2\\Users')
#	tree_id = tree_entry.tree_id
#	file_path = 'Administrator\\Desktop\\smb_test\\testfile1.txt'
#	
#	desired_access = FileAccessMask.FILE_READ_DATA
#	share_mode = ShareAccess.FILE_SHARE_READ
#	create_options = CreateOptions.FILE_NON_DIRECTORY_FILE
#	file_attrs = 0
#	create_disposition = CreateDisposition.FILE_OPEN
#	
#	file_id = await connection.create(tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs)
#	
#	await connection.query_info(tree_id, file_id)
#	await connection.read(tree_id, file_id, offset = 0, length = 20)
#	
#	tree_entry = await connection.tree_connect('\\\\10.10.10.2\\Users')
#	tree_id = tree_entry.tree_id
#	file_path = 'Administrator\\Desktop\\smb_test\\'
#	
#	desired_access = FileAccessMask.FILE_READ_DATA
#	share_mode = ShareAccess.FILE_SHARE_READ
#	create_options = CreateOptions.FILE_DIRECTORY_FILE
#	file_attrs = 0
#	create_disposition = CreateDisposition.FILE_OPEN
#	
#	file_id = await connection.create(tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs)
#	info = await connection.query_directory(tree_id, file_id)
#	print(str(info))
#	
#async def test_high(target):
#	#setting up NTLM auth
#	template_name = 'Windows10_15063_knowkey'
#	credential = Credential()
#	credential.username = 'victim'
#	credential.password = 'Passw0rd!1'
#	credential.domain = 'TEST'
#	
#	settings = NTLMHandlerSettings(credential, mode = 'CLIENT', template_name = template_name)
#	handler = NTLMAUTHHandler(settings)
#	
#	#setting up SPNEGO
#	spneg = SPNEGO()
#	spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
#	connection = SMBConnection(spneg, [NegotiateDialects.SMB210])
#	await connection.connect(target)
#	await connection.negotiate()
#	await connection.session_setup()
#	
#	end = SMBEndpoint()
#	end.connection = connection
#	
#	await end.test()
#	
#async def test_kerberos(target):
#	settings = {
#		'mode' : 'CLIENT',
#		'connection_string' : 'TEST/victim/pass:Passw0rd!1@10.10.10.2',
#		'target_string': 'cifs/WIN2019AD@TEST.CORP',
#		'dc_ip' : '10.10.10.2',
#	}
#	
#	handler = SMBKerberos(settings)
#	
#	#setting up SPNEGO
#	spneg = SPNEGO()
#	spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
#	connection = SMBConnection(spneg, [NegotiateDialects.SMB210])
#	await connection.connect(target)
#	await connection.negotiate()
#	await connection.session_setup()
#	tree_entry = await connection.tree_connect('\\\\10.10.10.2\\Users')
#	tree_id = tree_entry.tree_id
#	file_path = 'Administrator\\Desktop\\smb_test\\testfile1.txt'
#	
#	desired_access = FileAccessMask.FILE_READ_DATA
#	share_mode = ShareAccess.FILE_SHARE_READ
#	create_options = CreateOptions.FILE_NON_DIRECTORY_FILE
#	file_attrs = 0
#	create_disposition = CreateDisposition.FILE_OPEN
#	
#	file_id = await connection.create(tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs)
#
#async def test_sspi_kerberos(target):
#	settings = {
#		'mode' : 'CLIENT',
#		'username' : None,
#		'password' : None,
#		'target' : 'WIN2019AD',
#	}
#	handler = SMBKerberosSSPI(settings)
#	#setting up SPNEGO
#	spneg = SPNEGO()
#	spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
#	connection = SMBConnection(spneg, [NegotiateDialects.SMB210])
#	await connection.connect(target)
#	await connection.negotiate()
#	await connection.session_setup()
#	tree_entry = await connection.tree_connect('\\\\10.10.10.2\\Users')
#	tree_id = tree_entry.tree_id
#	file_path = 'Administrator\\Desktop\\smb_test\\testfile1.txt'
#	
#	desired_access = FileAccessMask.FILE_READ_DATA
#	share_mode = ShareAccess.FILE_SHARE_READ
#	create_options = CreateOptions.FILE_NON_DIRECTORY_FILE
#	file_attrs = 0
#	create_disposition = CreateDisposition.FILE_OPEN
#	
#	file_id = await connection.create(tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs)
#	
#async def test_sspi_ntlm(target):
#	settings = {
#		'mode' : 'CLIENT',
#	}
#	handler = SMBNTLMSSPI(settings)
#	#setting up SPNEGO
#	spneg = SPNEGO()
#	spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
#	connection = SMBConnection(spneg, [NegotiateDialects.SMB210])
#	await connection.connect(target)
#	await connection.negotiate()
#	await connection.session_setup()
#	tree_entry = await connection.tree_connect('\\\\10.10.10.2\\Users')
#	tree_id = tree_entry.tree_id
#	file_path = 'Administrator\\Desktop\\smb_test\\testfile1.txt'
#	
#	desired_access = FileAccessMask.FILE_READ_DATA
#	share_mode = ShareAccess.FILE_SHARE_READ
#	create_options = CreateOptions.FILE_NON_DIRECTORY_FILE
#	file_attrs = 0
#	create_disposition = CreateDisposition.FILE_OPEN
#	
#	file_id = await connection.create(tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs)
#	
#	await connection.terminate()
#	
#async def connection_test(target):
#	#setting up NTLM auth
#	template_name = 'Windows10_15063_knowkey'
#	credential = Credential()
#	credential.username = 'victim'
#	credential.password = 'Passw0rd!1'
#	credential.domain = 'TEST'
#	
#	settings = NTLMHandlerSettings(credential, mode = 'CLIENT', template_name = template_name)
#	handler = NTLMAUTHHandler(settings)
#	
#	#setting up SPNEGO
#	spneg = SPNEGO()
#	spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
#	connection = SMBConnection(spneg, [NegotiateDialects.SMB210])
#	await connection.connect(target)
#	await connection.negotiate()
#	await connection.session_setup()
#	#input(connection.get_extra_info())
#	tree_entry = await connection.tree_connect('\\\\10.10.10.2\\Users')
#	tree_id = tree_entry.tree_id
#	file_path = 'Administrator\\Desktop\\smb_test\\testfile1.txt'
#	
#	desired_access = FileAccessMask.FILE_READ_DATA
#	share_mode = ShareAccess.FILE_SHARE_READ
#	create_options = CreateOptions.FILE_NON_DIRECTORY_FILE
#	file_attrs = 0
#	create_disposition = CreateDisposition.FILE_OPEN
#	
#	file_id = await connection.create(tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs)
#	
#	await connection.query_info(tree_id, file_id)
#	data, remaining = await connection.read(tree_id, file_id, offset = 100, length = 60000)
#	
#	print(data)
#	print(remaining)
#	
#	tree_entry = await connection.tree_connect('\\\\10.10.10.2\\Users')
#	tree_id = tree_entry.tree_id
#	file_path = 'Administrator\\Desktop\\smb_test\\'
#	
#	desired_access = FileAccessMask.FILE_READ_DATA
#	share_mode = ShareAccess.FILE_SHARE_READ
#	create_options = CreateOptions.FILE_DIRECTORY_FILE
#	file_attrs = 0
#	create_disposition = CreateDisposition.FILE_OPEN
#	
#	file_id = await connection.create(tree_id, file_path, desired_access, share_mode, create_options, create_disposition, file_attrs)
#	info = await connection.query_directory(tree_id, file_id)
#	await connection.terminate()
#	print(str(info))
#	
#async def filereader_test(target):
#	#setting up NTLM auth
#	template_name = 'Windows10_15063_knowkey'
#	credential = Credential()
#	credential.username = 'victim'
#	credential.password = 'Passw0rd!1'
#	credential.domain = 'TEST'
#	
#	settings = NTLMHandlerSettings(credential, mode = 'CLIENT', template_name = template_name)
#	handler = NTLMAUTHHandler(settings)
#	
#	#setting up SPNEGO
#	spneg = SPNEGO()
#	spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
#	async with SMBConnection(spneg, target) as connection: 
#		await connection.login()
#		
#		async with SMBFileReader(connection) as reader:
#			await reader.open('\\\\10.10.10.2\\Users\\Administrator\\Desktop\\smb_test\\testfile1.txt')
#			data = await reader.read()
#			print(data)
#			await reader.seek(0,0)
#			data = await reader.read()
#			print(data)
#			await reader.seek(10,0)
#			data = await reader.read()
#			print(data)
#			await reader.seek(10,0)
#			data = await reader.read(5)
#			print(data)
#			await reader.seek(-10,2)
#			data = await reader.read(5)
#			print(data)
#	
#	
#			
#if __name__ == '__main__':
#	target = SMBTarget()
#	target.ip = '10.10.10.2'
#	target.port = 445
#	
#	target_bad = SMBTarget()
#	target_bad.ip = '10.10.10.66'
#	target_bad.port = 445
#
#	#asyncio.run(test(target))
#	#asyncio.run(test_kerberos(target))
#	#asyncio.run(test_sspi_kerberos(target))
#	#asyncio.run(test_sspi_ntlm(target))
#	#asyncio.run(connection_test(target))
#	#asyncio.run(test_high(target))
#	asyncio.run(filereader_test(target))
#
async def ctest(cu):
	conn = cu.get_connection()
	await conn.login()
	await conn.ghosting()
			
		


if __name__ == '__main__':
	from aiosmb.commons.connection.url import SMBConnectionURL

	logger.setLevel(2)
	#url = 'smb+ntlm-password://TEST\\victim:Passw0rd!1@10.10.10.2'
	url = 'smb222+ntlm-password://TEST\\victim:Passw0rd!1@10.10.10.2'
	#url = 'smb202+ntlm-password://work\\work:work@10.10.10.103'
	#url = 'smb+ntlm-password://smbtest\\smbtest:smbtest@10.200.200.154'
	cu = SMBConnectionURL(url)
	
	asyncio.run(ctest(cu))
	


import importlib
import importlib.util

#
##key = name of the cipher, value=list of module names, in order of preference
#preftable = {
#	'DES' : ['pyCrypto','pure'], 
#	'TDES': ['pyCrypto','pure'], 
#	'AES' : ['cryptography','pyCrypto','pure'], 
#	'RC4' : ['cryptography','pyCrypto','pure'],
#
#}
#
#available_modules = ['pure']
#
#if importlib.util.find_spec("cryptography") is not None:
#	#print('Found cryptography package!')
#	available_modules.append("cryptography")
#
#elif importlib.util.find_spec("pyCrypto") is not None:
#	#print('Found cryptography package!')
#	available_modules.append("pyCrypto")
#
#
##https://stackoverflow.com/questions/8790003/dynamically-import-a-method-in-a-file-from-a-string
#def import_from(module, name):
#	module = __import__(module, fromlist=[name])
#	return getattr(module, name)
#
#
#def getPreferredCipher(cipherName):
#	if cipherName not in preftable:
#		raise Exception('Cipher %s doesnt have any preferences set!' % cipherName)
#	possible_prefmodule = list(set(preftable[cipherName]).intersection(set(available_modules)))
#	selected_module = None
#	for moduleName in preftable[cipherName]:
#		if moduleName in possible_prefmodule:
#			selected_module = moduleName
#
#	if selected_module is None:
#		raise Exception('Could not find any modules to load cipher %s' % cipherName)
#
#
#	#print('Preferred module selected for cipher %s is %s' % (cipherName, selected_module))
#	moduleName = 'aiosmb.crypto.%s' % cipherName
#	objectName = selected_module + cipherName
#	return import_from(moduleName , objectName)
#
#def getSpecificCipher(cipherName, moduleBaseName):
#	moduleName = 'aiosmb.crypto.%s' % cipherName
#	objectName = '%s%s' % (moduleBaseName, cipherName)
#	return import_from(moduleName , objectName)


#import ciphers
# TODO: fix the dynamic imports, currently only supporting pure-python ciphers for two reasons:
# 1. dynamic import messes up some scripts like pyinstaller/nuitka/py2exe
# 2. additional effort needed to support more crypto libs anyhow

from aiosmb.crypto.AES import pureAES
from aiosmb.crypto.RC4 import pureRC4
from aiosmb.crypto.DES import pureDES

DES  = pureDES #getPreferredCipher('DES')
AES  = pureAES #getPreferredCipher('AES')
RC4  = pureRC4 #getPreferredCipher('RC4')
#TDES = getPreferredCipher('TDES')




#
#
#
#
#
#


import enum
import asyncio
import ipaddress

from aiosmb import logger
from aiosmb.commons.exceptions import *

from asysocks.client import SOCKSClient
from asysocks.common.comms import SocksQueueComms


class SocksProxyConnection:
	"""
	Generic asynchronous TCP socket class, nothing SMB related.
	Creates the connection and channels incoming/outgoing bytes via asynchonous queues.
	"""
	def __init__(self, target = None, socket = None):
		self.target = target
		self.socket = socket #for future, if we want a custom soscket
		
		self.client = None
		self.proxy_task = None

		self.out_queue = None#asyncio.Queue()
		self.in_queue = None#asyncio.Queue()
		
	async def disconnect(self):
		"""
		Disconnects from the socket.
		Stops the reader and writer streams.
		"""
		self.proxy_task.cancel()		
		
	async def connect(self):
		"""
		
		"""
		
		self.out_queue = asyncio.Queue()
		self.in_queue = asyncio.Queue()
		comms = SocksQueueComms(self.out_queue, self.in_queue)

		self.target.proxy.target.endpoint_ip = self.target.ip
		self.target.proxy.target.endpoint_port = int(self.target.port)

		self.client = SOCKSClient(comms, self.target.proxy.target, self.target.proxy.auth)
		self.proxy_task = asyncio.create_task(self.client.run())
		return

			

			


			